@@ -19,7 +19,6 @@ import unittest
 import mock
 from oslo_serialization import jsonutils
 from oslo_utils import uuidutils
-from requests import exceptions as requests_exceptions
 from requests import models
 
 from vmware_nsxlib import v3
@@ -27,6 +26,7 @@ from vmware_nsxlib.v3 import client as nsx_client
 from vmware_nsxlib.v3 import client_cert
 from vmware_nsxlib.v3 import cluster as nsx_cluster
 from vmware_nsxlib.v3 import config
+from vmware_nsxlib.v3 import nsx_constants
 from vmware_nsxlib.v3 import utils
 
 NSX_USER = 'admin'
@@ -50,23 +50,13 @@ DNS_NAMESERVERS = ['1.1.1.1']
 DNS_DOMAIN = 'openstacklocal'
 
 JSESSIONID = 'my_sess_id'
+LATEST_VERSION = nsx_constants.NSX_VERSION_3_1_0
 
 
 def _mock_nsxlib():
     def _return_id_key(*args, **kwargs):
         return {'id': uuidutils.generate_uuid()}
 
-    def _mock_add_rules_in_section(*args):
-        # NOTE(arosen): the code in the neutron plugin expects the
-        # neutron rule id as the display_name.
-        rules = args[0]
-        return {
-            'rules': [
-                {'display_name': rule['display_name'],
-                 'id': uuidutils.generate_uuid()}
-                for rule in rules
-            ]}
-
     def _mock_limits(*args):
         return utils.TagLimits(20, 40, 15)
 
@@ -90,10 +80,6 @@ def _mock_nsxlib():
     mocking.append(mock.patch(
         "vmware_nsxlib.v3.security.NsxLibNsGroup.list"))
 
-    mocking.append(mock.patch(
-        "vmware_nsxlib.v3.security.NsxLibFirewallSection.add_rules",
-        side_effect=_mock_add_rules_in_section))
-
     mocking.append(mock.patch(
         ("vmware_nsxlib.v3.core_resources."
          "NsxLibTransportZone.get_id_by_name_or_id"),
@@ -157,7 +143,10 @@ class NsxLibTestCase(unittest.TestCase):
 
     def setUp(self, *args, **kwargs):
         super(NsxLibTestCase, self).setUp()
-        self.mocking = _mock_nsxlib()
+        if kwargs.get('with_mocks', True):
+            self.mocking = _mock_nsxlib()
+        else:
+            self.mocking = []
 
         if self.use_client_cert_auth():
             nsxlib_config = get_nsxlib_config_with_client_cert()
@@ -190,18 +179,12 @@ class MemoryMockAPIProvider(nsx_cluster.AbstractHTTPProvider):
         return "Memory mock API"
 
     def validate_connection(self, cluster_api, endpoint, conn):
-        return
+        return True
 
     def new_connection(self, cluster_api, provider):
         # all callers use the same backing
         return self._store
 
-    def is_connection_exception(self, exception):
-        return isinstance(exception, requests_exceptions.ConnectionError)
-
-    def is_timeout_exception(self, exception):
-        return isinstance(exception, requests_exceptions.Timeout)
-
 
 class NsxClientTestCase(NsxLibTestCase):
 
@@ -239,7 +222,8 @@ class NsxClientTestCase(NsxLibTestCase):
                 nsx_api_managers=nsx_api_managers or [NSX_MANAGER],
                 plugin_scope=PLUGIN_SCOPE,
                 plugin_tag=PLUGIN_TAG,
-                plugin_ver=PLUGIN_VER)
+                plugin_ver=PLUGIN_VER,
+                cluster_unavailable_retry=True)
 
             super(NsxClientTestCase.MockNSXClusteredAPI, self).__init__(
                 nsxlib_config)
@@ -294,7 +278,7 @@ class NsxClientTestCase(NsxLibTestCase):
             super(NsxClientTestCase.MockHTTPProvider, self).__init__()
             if isinstance(session_response, list):
                 self._session_responses = session_response
-            elif session_response:
+            elif session_response is not None:
                 self._session_responses = [session_response]
             else:
                 self._session_responses = None
@@ -349,6 +333,7 @@ class NsxClientTestCase(NsxLibTestCase):
 
         def validate_connection(self, cluster_api, endpoint, conn):
             assert conn is not None
+            return True
 
     def mock_nsx_clustered_api(self, session_response=None, **kwargs):
         orig_request = nsx_cluster.TimeoutSession.request
@@ -409,12 +394,14 @@ class NsxClientTestCase(NsxLibTestCase):
         return client
 
     def new_mocked_cluster(self, conf_managers, validate_conn_func,
-                           concurrent_connections=None):
+                           concurrent_connections=None, exceptions=None):
         mock_provider = mock.Mock()
         mock_provider.default_scheme = 'https'
         mock_provider.validate_connection = validate_conn_func
 
         nsxlib_config = get_default_nsxlib_config()
+        if exceptions:
+            nsxlib_config.exception_config = exceptions
         if concurrent_connections:
             nsxlib_config.concurrent_connections = concurrent_connections
         nsxlib_config.http_provider = mock_provider
