@@ -16,6 +16,7 @@
 
 import mock
 
+from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
 from vmware_nsxlib.tests.unit.v3.policy import test_resources
 from vmware_nsxlib.v3 import exceptions as nsxlib_exc
 from vmware_nsxlib.v3.policy import constants
@@ -847,14 +848,25 @@ class TestPolicyLBService(test_resources.NsxPolicyLibTestCase):
                 tenant=TEST_TENANT)
             self.assert_called_with_def(api_call, expected_def)
 
-    def test_list(self):
+    def _test_list(self, silent=False, silent_if_empty=False):
+        s1 = {'id': 'xxx', 'display_name': 'yyy'}
         with mock.patch.object(self.policy_api, "list",
-                               return_value={'results': []}) as api_call:
-            result = self.resourceApi.list(tenant=TEST_TENANT)
+                               return_value={'results': [s1]}) as api_call:
+            result = self.resourceApi.list(tenant=TEST_TENANT, silent=silent,
+                                           silent_if_empty=silent_if_empty)
             expected_def = lb_defs.LBServiceDef(
                 tenant=TEST_TENANT)
             self.assert_called_with_def(api_call, expected_def)
-            self.assertEqual([], result)
+            self.assertEqual([s1], result)
+
+    def test_list(self):
+        self._test_list()
+
+    def test_list_total_silence(self):
+        self._test_list(silent=True)
+
+    def test_list_silent_if_empty(self):
+        self._test_list(silent_if_empty=True)
 
     def test_update(self):
         obj_id = '111'
@@ -884,6 +896,25 @@ class TestPolicyLBService(test_resources.NsxPolicyLibTestCase):
                 relax_scale_validation=relax_scale_validation)
             self.assert_called_with_def(update_call, expected_def)
 
+    def test_update_customized(self):
+        obj_id = '111'
+        name = 'name'
+        tags = [{'tag': '1', 'scope': '2'}]
+
+        def update_callback(body):
+            body['tags'] = tags
+
+        with self.mock_get(obj_id, name), \
+            mock.patch.object(self.policy_api.client, "update") as update_call:
+            self.resourceApi.update_customized(
+                obj_id, update_callback)
+
+            update_call.assert_called_once_with(
+                'infra/lb-services/%s' % obj_id,
+                {'id': obj_id, 'display_name': name,
+                 'resource_type': 'LBService',
+                 'tags': tags})
+
     def test_get_status(self):
         obj_id = '111'
         with mock.patch.object(self.policy_api, "get") as api_call:
@@ -1103,7 +1134,8 @@ class TestPolicyLBVirtualServer(test_resources.NsxPolicyLibTestCase):
                 virtual_server_id=obj_id, name=vs_name,
                 max_concurrent_connections=None,
                 tenant=TEST_TENANT)
-            update_call.assert_called_with(mock.ANY, partial_updates=False)
+            update_call.assert_called_with(mock.ANY, partial_updates=False,
+                                           force=False)
             self.assert_called_with_def(update_call, expected_def)
 
     def test_add_lb_rule(self):
@@ -1275,6 +1307,36 @@ class TestPolicyLBVirtualServer(test_resources.NsxPolicyLibTestCase):
                        {'display_name': 'xx', 'actions': '22'}])
             self.assert_called_with_def(update_call, expected_def)
 
+    def test_update_lb_rule_suffix(self):
+        vs_obj_id = '111'
+        vs_name = 'name-name'
+        with self.mock_get(
+                vs_obj_id, vs_name,
+                rules=[{'display_name': 'xx_with_suffix', 'actions': '11'},
+                       {'display_name': 'yy'}]), \
+            self.mock_create_update() as update_call:
+            self.resourceApi.update_lb_rule(
+                vs_obj_id, 'xx', actions='22',
+                compare_name_suffix='suffix')
+
+            expected_def = lb_defs.LBVirtualServerDef(
+                virtual_server_id=vs_obj_id,
+                rules=[{'display_name': 'xx', 'actions': '22'},
+                       {'display_name': 'yy'}])
+            self.assert_called_with_def(update_call, expected_def)
+
+    def test_update_lb_rule_wrong_suffix(self):
+        vs_obj_id = '111'
+        vs_name = 'name-name'
+        with self.mock_get(
+                vs_obj_id, vs_name,
+                rules=[{'display_name': 'xx_with_suffix', 'actions': '11'},
+                       {'display_name': 'yy'}]):
+            self.assertRaises(nsxlib_exc.ResourceNotFound,
+                              self.resourceApi.update_lb_rule,
+                              vs_obj_id, 'xx', actions='22',
+                              compare_name_suffix='bad')
+
     def test_remove_lb_rule(self):
         vs_obj_id = '111'
         vs_name = 'name-name'
@@ -1289,6 +1351,37 @@ class TestPolicyLBVirtualServer(test_resources.NsxPolicyLibTestCase):
                 rules=[{'display_name': 'yy'}])
             self.assert_called_with_def(update_call, expected_def)
 
+    def test_remove_lb_rule_by_suffix(self):
+        vs_obj_id = '111'
+        vs_name = 'name-name'
+        with self.mock_get(vs_obj_id, vs_name,
+                           rules=[{'display_name': 'xx_with_suffix'},
+                                  {'display_name': 'yy'}]), \
+            self.mock_create_update() as update_call:
+            self.resourceApi.remove_lb_rule(vs_obj_id, 'with_suffix',
+                                            check_name_suffix=True)
+
+            expected_def = lb_defs.LBVirtualServerDef(
+                virtual_server_id=vs_obj_id,
+                rules=[{'display_name': 'yy'}])
+            self.assert_called_with_def(update_call, expected_def)
+
+    def test_remove_lb_rule_wrong_suffix(self):
+        vs_obj_id = '111'
+        vs_name = 'name-name'
+        with self.mock_get(vs_obj_id, vs_name,
+                           rules=[{'display_name': 'xx_with_suffix'},
+                                  {'display_name': 'yy'}]),\
+            self.mock_create_update() as update_call:
+            self.resourceApi.remove_lb_rule(vs_obj_id, 'wrong_suffiX',
+                                            check_name_suffix=True)
+
+            expected_def = lb_defs.LBVirtualServerDef(
+                virtual_server_id=vs_obj_id,
+                rules=[{'display_name': 'xx_with_suffix'},
+                       {'display_name': 'yy'}])
+            self.assert_called_with_def(update_call, expected_def)
+
     def test_build_access_list_control(self):
         lb_acl = self.resourceApi.build_access_list_control(
             constants.ACTION_ALLOW, 'fake_group_path', True)
@@ -1340,6 +1433,18 @@ class TestPolicyLBVirtualServer(test_resources.NsxPolicyLibTestCase):
                 virtual_server_id=vs_id, name=vs_name)
             self.assert_called_with_def(update_call, expected_def)
 
+    def test_remove_dlb_virtual_server_persistence_profile(self):
+        vs_id = 'test-id'
+        vs_name = 'test-name'
+        with self.mock_get(
+                vs_id, vs_name, lb_persistence_profile_path='test-profile'), \
+                self.mock_create_update() as update_call:
+            self.resourceApi.remove_dlb_virtual_server_persistence_profile(
+                vs_id)
+            expected_def = lb_defs.LBVirtualServerDef(
+                virtual_server_id=vs_id, name=vs_name)
+            self.assert_called_with_def(update_call, expected_def)
+
 
 class TestPolicyLBPoolApi(test_resources.NsxPolicyLibTestCase):
 
@@ -1398,6 +1503,60 @@ class TestPolicyLBPoolApi(test_resources.NsxPolicyLibTestCase):
             self.assert_called_with_def(api_call, expected_def)
             self.assertIsNotNone(result)
 
+    def test_create_with_retry_stale_revision(self):
+        name = 'd1'
+        description = 'desc'
+        obj_id = '111'
+        members = [
+            lb_defs.LBPoolMemberDef(ip_address='10.0.0.1')]
+        algorithm = 'algo'
+        active_monitor_paths = 'path1'
+        member_group = 'group1'
+        snat_translation = False
+        with mock.patch.object(self.policy_api, "create_or_update",
+                               side_effect=nsxlib_exc.StaleRevision
+                               ) as api_call:
+            with self.assertRaises(nsxlib_exc.StaleRevision):
+                self.resourceApi.create_or_overwrite(
+                    name,
+                    lb_pool_id=obj_id,
+                    description=description,
+                    members=members,
+                    active_monitor_paths=active_monitor_paths,
+                    algorithm=algorithm,
+                    member_group=member_group,
+                    snat_translation=snat_translation,
+                    tenant=TEST_TENANT)
+                self.assertEqual(nsxlib_testcase.NSX_MAX_ATTEMPTS,
+                                 api_call.call_count)
+
+    def test_create_with_retry_pending_delete(self):
+        name = 'd1'
+        description = 'desc'
+        obj_id = '111'
+        members = [
+            lb_defs.LBPoolMemberDef(ip_address='10.0.0.1')]
+        algorithm = 'algo'
+        active_monitor_paths = 'path1'
+        member_group = 'group1'
+        snat_translation = False
+        with mock.patch.object(self.policy_api, "create_or_update",
+                               side_effect=nsxlib_exc.NsxPendingDelete
+                               ) as api_call:
+            with self.assertRaises(nsxlib_exc.NsxPendingDelete):
+                self.resourceApi.create_or_overwrite(
+                    name,
+                    lb_pool_id=obj_id,
+                    description=description,
+                    members=members,
+                    active_monitor_paths=active_monitor_paths,
+                    algorithm=algorithm,
+                    member_group=member_group,
+                    snat_translation=snat_translation,
+                    tenant=TEST_TENANT)
+                self.assertEqual(nsxlib_testcase.NSX_MAX_ATTEMPTS,
+                                 api_call.call_count)
+
     def test_delete(self):
         obj_id = '111'
         with mock.patch.object(self.policy_api, "delete") as api_call:
@@ -1472,6 +1631,43 @@ class TestPolicyLBPoolApi(test_resources.NsxPolicyLibTestCase):
                 tenant=TEST_TENANT)
             self.assert_called_with_def(update_call, expected_def)
 
+    def test_update_without_partial_patch(self):
+        obj_id = '111'
+        name = 'new name'
+        description = 'new desc'
+        members = [{'ip_address': '10.0.0.1'}]
+        algorithm = 'algo'
+        active_monitor_paths = ['path1']
+        member_group = 'group1'
+        snat_translation = False
+        with mock.patch.object(self.policy_api, "get",
+                               return_value={'id': obj_id}), \
+            mock.patch.object(self.policy_api,
+                              "create_or_update") as update_call:
+            self.resourceApi.update(obj_id,
+                                    name=name,
+                                    description=description,
+                                    members=members,
+                                    active_monitor_paths=active_monitor_paths,
+                                    algorithm=algorithm,
+                                    member_group=member_group,
+                                    snat_translation=snat_translation,
+                                    tenant=TEST_TENANT,
+                                    allow_partial_updates=False)
+            expected_def = lb_defs.LBPoolDef(
+                lb_pool_id=obj_id,
+                name=name,
+                description=description,
+                members=members,
+                active_monitor_paths=active_monitor_paths,
+                algorithm=algorithm,
+                member_group=member_group,
+                snat_translation=snat_translation,
+                tenant=TEST_TENANT)
+            update_call.assert_called_with(mock.ANY, partial_updates=False,
+                                           force=False)
+            self.assert_called_with_def(update_call, expected_def)
+
     def test_add_monitor_to_pool(self):
         obj_id = '111'
         active_monitor_paths = ['path1']
