@@ -260,11 +260,13 @@ class TestPolicyTransaction(policy_testcase.TestPolicyApi):
         nat_rule1 = {"action": constants.NAT_ACTION_SNAT,
                      "display_name": "snat rule",
                      "id": nat_rule_id1,
-                     "resource_type": "PolicyNatRule"}
+                     "resource_type": "PolicyNatRule",
+                     "firewall_match": constants.NAT_FIREWALL_MATCH_BYPASS}
         nat_rule2 = {"action": constants.NAT_ACTION_DNAT,
                      "display_name": "dnat rule",
                      "id": nat_rule_id2,
-                     "resource_type": "PolicyNatRule"}
+                     "resource_type": "PolicyNatRule",
+                     "firewall_match": constants.NAT_FIREWALL_MATCH_BYPASS}
 
         policy_nat = {"id": "USER",
                       "resource_type": "PolicyNat",
@@ -391,7 +393,8 @@ class TestPolicyTransaction(policy_testcase.TestPolicyApi):
         self.assert_infra_patch_call(expected_body)
 
     @mock.patch('vmware_nsxlib.v3.policy.core_defs.NsxPolicyApi.get')
-    def test_updating_security_policy_and_dfw_rules(self, mock_get_api):
+    def _test_updating_security_policy_and_dfw_rules(
+        self, use_child_rules, mock_get_api):
         dfw_rule1 = {'id': 'rule_id1', 'action': 'ALLOW',
                      'display_name': 'rule1', 'description': None,
                      'direction': 'IN_OUT', 'ip_protocol': 'IPV4_IPV6',
@@ -449,15 +452,20 @@ class TestPolicyTransaction(policy_testcase.TestPolicyApi):
                 name=security_policy['display_name'],
                 domain_id=domain_id,
                 map_id=map_id,
-                entries=dfw_rule_entries
+                entries=dfw_rule_entries,
+                use_child_rules=use_child_rules
             )
 
         dfw_rule1['display_name'] = new_rule_name
         dfw_rule1['direction'] = new_direction
-        child_rules = [{'resource_type': 'ChildRule', 'Rule': dfw_rule1},
-                       {'resource_type': 'ChildRule', 'Rule': dfw_rule2,
-                        'marked_for_delete': True}]
-        security_policy.update({'children': child_rules})
+        if use_child_rules:
+            child_rules = [{'resource_type': 'ChildRule', 'Rule': dfw_rule1},
+                           {'resource_type': 'ChildRule', 'Rule': dfw_rule2,
+                            'marked_for_delete': True}]
+            security_policy.update({'children': child_rules})
+        else:
+            security_policy['rules'] = copy.deepcopy([dfw_rule1, dfw_rule2])
+
         child_security_policies = [{
             'resource_type': 'ChildSecurityPolicy',
             'SecurityPolicy': security_policy
@@ -471,6 +479,12 @@ class TestPolicyTransaction(policy_testcase.TestPolicyApi):
                          'children': child_domains}
         self.assert_infra_patch_call(expected_body)
 
+    def test_updating_security_policy_and_dfw_rules(self):
+        return self._test_updating_security_policy_and_dfw_rules(True)
+
+    def test_updating_security_policy_and_dfw_rules_no_child_rules(self):
+        return self._test_updating_security_policy_and_dfw_rules(False)
+
     @mock.patch('vmware_nsxlib.v3.policy.core_defs.NsxPolicyApi.get')
     def test_updating_security_policy_with_no_entries_set(self, mock_get_api):
         dfw_rule1 = {'id': 'rule_id1', 'action': 'ALLOW',
