@@ -16,27 +16,35 @@
 import unittest
 
 import mock
+from requests import codes
 from requests import exceptions as requests_exceptions
 from requests import models
 import six.moves.urllib.parse as urlparse
 
+from oslo_serialization import jsonutils
+
 from vmware_nsxlib.tests.unit.v3 import mocks
 from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
 from vmware_nsxlib import v3
 from vmware_nsxlib.v3 import client
 from vmware_nsxlib.v3 import client_cert
 from vmware_nsxlib.v3 import cluster
+from vmware_nsxlib.v3 import config
 from vmware_nsxlib.v3 import exceptions as nsxlib_exc
 
 
 def _validate_conn_up(*args, **kwargs):
-    return
+    return True
 
 
 def _validate_conn_down(*args, **kwargs):
     raise requests_exceptions.ConnectionError()
 
 
+def _validate_server_busy(*args, **kwargs):
+    raise nsxlib_exc.ServerBusy(details="Test")
+
+
 def get_sess_create_resp():
     sess_create_response = models.Response()
     sess_create_response.status_code = 200
@@ -171,7 +179,6 @@ class RequestsHTTPProviderTestCase(unittest.TestCase):
         mock_cluster = mock.Mock()
         mock_cluster.nsxlib_config = mock.Mock()
         mock_cluster.nsxlib_config.url_base = 'abc'
-        mock_cluster.nsxlib_config.keepalive_section = 'transport-zones'
         provider = cluster.NSXRequestsHTTPProvider()
 
         with mock.patch.object(client.JSONRESTClient, "get",
@@ -181,7 +188,7 @@ class RequestsHTTPProviderTestCase(unittest.TestCase):
                               mock_cluster, mock_ep, mock_conn)
 
         with mock.patch.object(client.JSONRESTClient, "get",
-                               return_value={'result_count': 1}):
+                               return_value={'healthy': True}):
             provider.validate_connection(mock_cluster, mock_ep, mock_conn)
 
     def _validate_con_mocks(self, nsx_version):
@@ -194,26 +201,11 @@ class RequestsHTTPProviderTestCase(unittest.TestCase):
         mock_ep.provider.url = 'https://1.2.3.4'
         conf = mock.Mock()
         conf.url_base = 'abc'
-        conf.keepalive_section = 'transport-zones'
-        conf.validate_connection_method = nsxlib.validate_connection_method
         mock_cluster = mock.Mock()
         mock_cluster.nsxlib_config = conf
         return (mock_cluster, mock_ep, mock_conn)
 
-    def test_validate_connection_method_v1(self):
-        mock_cluster, mock_ep, mock_conn = self._validate_con_mocks('2.3.0')
-        provider = cluster.NSXRequestsHTTPProvider()
-        with mock.patch.object(client.JSONRESTClient, "get",
-                               return_value={'application_status': 'DOWN'}):
-            self.assertRaises(nsxlib_exc.ResourceNotFound,
-                              provider.validate_connection,
-                              mock_cluster, mock_ep, mock_conn)
-
-        with mock.patch.object(client.JSONRESTClient, "get",
-                               return_value={'application_status': 'WORKING'}):
-            provider.validate_connection(mock_cluster, mock_ep, mock_conn)
-
-    def test_validate_connection_method_v2(self):
+    def test_validate_connection_method(self):
         mock_cluster, mock_ep, mock_conn = self._validate_con_mocks('2.4.0')
         provider = cluster.NSXRequestsHTTPProvider()
         with mock.patch.object(client.JSONRESTClient, "get",
@@ -298,9 +290,10 @@ class NsxV3ClusteredAPIWithClientCertTestCase(NsxV3ClusteredAPITestCase):
 
 class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
 
-    def _test_health(self, validate_fn, expected_health):
+    def _test_health(self, validate_fn, expected_health, exceptions=None):
         conf_managers = ['8.9.10.11', '9.10.11.12']
-        api = self.new_mocked_cluster(conf_managers, validate_fn)
+        api = self.new_mocked_cluster(conf_managers, validate_fn,
+                                      exceptions=exceptions)
 
         self.assertEqual(expected_health, api.health)
 
@@ -310,6 +303,8 @@ class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
             if endpoint.provider.id == '8.9.10.11':
                 raise Exception()
 
+            return True
+
         self._test_health(_validate, cluster.ClusterHealth.ORANGE)
 
     def test_green_health(self):
@@ -326,6 +321,16 @@ class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
         self.assertRaises(nsxlib_exc.ServiceClusterUnavailable,
                           api.get, 'api/v1/transport-zones')
 
+    def test_cluster_validate_with_exception_retry(self):
+        self._test_health(_validate_server_busy, cluster.ClusterHealth.GREEN)
+
+    def test_cluster_validate_with_exception_no_retry(self):
+        exceptions = config.ExceptionConfig()
+        exceptions.retriables = []
+        self._test_health(_validate_server_busy,
+                          cluster.ClusterHealth.RED,
+                          exceptions)
+
     def test_cluster_proxy_stale_revision(self):
 
         def stale_revision():
@@ -336,6 +341,112 @@ class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
         self.assertRaises(nsxlib_exc.StaleRevision,
                           api.get, 'api/v1/transport-zones')
 
+    def test_cluster_down_and_retry_on_server_error(self):
+
+        def server_error():
+            return mocks.MockRequestsResponse(
+                codes.INTERNAL_SERVER_ERROR,
+                jsonutils.dumps({'error_message': 'test', 'error_code': 98}))
+
+        def all_well():
+            return mocks.MockRequestsResponse(
+                codes.OK,
+                jsonutils.dumps({'id': 'test'}))
+
+        conf_managers = ['8.9.10.11', '9.10.11.12', '10.11.12.13']
+        exceptions = config.ExceptionConfig()
+        exceptions.ground_triggers.append(nsxlib_exc.CannotConnectToServer)
+        api = self.mock_nsx_clustered_api(
+            nsx_api_managers=conf_managers,
+            session_response=[server_error, all_well])
+        api.nsxlib_config.exception_config = exceptions
+
+        api.get('api/v1/transport-zones')
+        # first manager should go down, second one is confirmed as UP
+        # after retry
+        self.assertEqual(cluster.ClusterHealth.ORANGE, api.health)
+
+    def test_max_retry_attempts_on_server_error(self):
+
+        def server_error():
+            return mocks.MockRequestsResponse(
+                codes.INTERNAL_SERVER_ERROR,
+                jsonutils.dumps({'error_message': 'test', 'error_code': 98}))
+
+        conf_managers = ['8.9.10.11', '9.10.11.12', '10.11.12.13']
+        exceptions = config.ExceptionConfig()
+        exceptions.ground_triggers.append(nsxlib_exc.CannotConnectToServer)
+        max_attempts = 4
+        api = self.mock_nsx_clustered_api(
+            nsx_api_managers=conf_managers,
+            max_attempts=max_attempts,
+            session_response=[server_error for i in range(0, max_attempts)])
+        api.nsxlib_config.cluster_unavailable_retry = False
+        api.nsxlib_config.exception_config = exceptions
+
+        self.assertRaises(nsxlib_exc.ServiceClusterUnavailable,
+                          api.get, 'api/v1/transport-zones')
+        self.assertEqual(cluster.ClusterHealth.RED, api.health)
+
+    def test_max_retry_attempts_on_retriable_error(self):
+
+        def server_error():
+            return mocks.MockRequestsResponse(
+                codes.INTERNAL_SERVER_ERROR,
+                jsonutils.dumps({'error_message': 'test', 'error_code': 98}))
+
+        conf_managers = ['8.9.10.11', '9.10.11.12', '10.11.12.13']
+        exceptions = config.ExceptionConfig()
+        max_attempts = 2
+        api = self.mock_nsx_clustered_api(
+            nsx_api_managers=conf_managers,
+            max_attempts=max_attempts,
+            session_response=[server_error for i in range(0, max_attempts)])
+        api.nsxlib_config.exception_config = exceptions
+
+        self.assertRaises(nsxlib_exc.CannotConnectToServer,
+                          api.get, 'api/v1/transport-zones')
+        # This exception does not ground endpoint
+        self.assertEqual(cluster.ClusterHealth.GREEN, api.health)
+
+    def test_regenerate_error(self):
+
+        def server_error():
+            return mocks.MockRequestsResponse(
+                codes.INTERNAL_SERVER_ERROR,
+                jsonutils.dumps({'error_message': 'test', 'error_code': 99}))
+
+        conf_managers = ['8.9.10.11', '9.10.11.12', '10.11.12.13']
+        exceptions = config.ExceptionConfig()
+        max_attempts = 2
+        api = self.mock_nsx_clustered_api(
+            nsx_api_managers=conf_managers,
+            max_attempts=max_attempts,
+            session_response=[server_error for i in range(0, max_attempts)])
+        api.nsxlib_config.exception_config = exceptions
+
+        self.assertRaises(nsxlib_exc.ClientCertificateNotTrusted,
+                          api.get, 'api/v1/transport-zones')
+        self.assertEqual(cluster.ClusterHealth.GREEN, api.health)
+
+    def test_exception_translation_on_ground_error(self):
+
+        def server_error():
+            raise requests_exceptions.ConnectionError()
+
+        conf_managers = ['8.9.10.11', '9.10.11.12', '10.11.12.13']
+        exceptions = config.ExceptionConfig()
+        max_attempts = 4
+        api = self.mock_nsx_clustered_api(
+            nsx_api_managers=conf_managers,
+            max_attempts=max_attempts,
+            session_response=[server_error for i in range(0, max_attempts)])
+        api.nsxlib_config.exception_config = exceptions
+
+        self.assertRaises(nsxlib_exc.ServiceClusterUnavailable,
+                          api.get, 'api/v1/transport-zones')
+        self.assertEqual(cluster.ClusterHealth.RED, api.health)
+
     def test_cluster_proxy_connection_establish_error(self):
 
         def connect_timeout():
@@ -390,6 +501,7 @@ class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
         max_attempts = 3
         api = self.mock_nsx_clustered_api(nsx_api_managers=conf_managers,
                                           max_attempts=max_attempts)
+        api.nsxlib_config.cluster_unavailable_retry = False
         api._validate = mock.Mock()
         eps = list(api._endpoints.values())
 
@@ -412,8 +524,7 @@ class ClusteredAPITestCase(nsxlib_testcase.NsxClientTestCase):
         # until retries have been exhausted
         api.nsxlib_config.cluster_unavailable_retry = True
         self.assertEqual(api._select_endpoint(), None)
-        self.assertEqual(api._validate.call_count,
-                         api.nsxlib_config.max_attempts * len(eps))
+        self.assertEqual(api._validate.call_count, len(eps))
 
         # simulate the case where 1 endpoint finally goes up
         self.validate_count = 0
