@@ -27,6 +27,9 @@ from vmware_nsxlib.v3 import nsx_constants as const
 class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
     """Tests for vmware_nsxlib.v3.security.NsxLibFirewallSection"""
 
+    def setUp(self, *args, **kwargs):
+        super(TestNsxLibFirewallSection, self).setUp(with_mocks=False)
+
     def test_get_logicalport_reference(self):
         mock_port = '3ed55c9f-f879-4048-bdd3-eded92465252'
         result = self.nsxlib.firewall_section.get_logicalport_reference(
@@ -60,6 +63,26 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
         }
         self.assertEqual(expected, result)
 
+    def test_get_rule_dict(self):
+        result = self.nsxlib.firewall_section.get_rule_dict(
+            'display_name', sources='sources', destinations='destinations',
+            direction=const.IN_OUT, ip_protocol=const.IPV4_IPV6,
+            services='services', action=const.FW_ACTION_ALLOW,
+            logged=True, disabled=True, applied_tos='applied_tos',
+            rule_tag='rule_tag')
+        expected = {'display_name': 'display_name',
+                    'sources': 'sources',
+                    'destinations': 'destinations',
+                    'direction': const.IN_OUT,
+                    'ip_protocol': const.IPV4_IPV6,
+                    'services': 'services',
+                    'action': const.FW_ACTION_ALLOW,
+                    'logged': True,
+                    'disabled': True,
+                    'applied_tos': 'applied_tos',
+                    'rule_tag': 'rule_tag'}
+        self.assertEqual(expected, result)
+
     def test_create_rules_with_protocol(self):
         with mock.patch("vmware_nsxlib.v3.security.NsxLibFirewallSection"
                         ".add_rules") as add_rules:
@@ -73,8 +96,8 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
             section_id = 'section-id'
             group_id = 'nsgroup-id'
             target_id = 'dummy'
-            self.nsxlib.firewall_section.create_rules(
-                None, section_id, group_id, False,
+            self.nsxlib.firewall_section.create_section_rules(
+                section_id, group_id, False,
                 "ALLOW", rules, {rule_id: target_id})
             add_rules.assert_called_once_with([
                 {'display_name': mock.ANY,
@@ -105,8 +128,8 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
             section_id = 'section-id'
             group_id = 'nsgroup-id'
             target_id = 'dummy'
-            self.nsxlib.firewall_section.create_rules(
-                None, section_id, group_id, False,
+            self.nsxlib.firewall_section.create_section_rules(
+                section_id, group_id, False,
                 "ALLOW", rules, {rule_id: target_id})
             add_rules.assert_called_once_with([
                 {'display_name': mock.ANY,
@@ -139,8 +162,8 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
             section_id = 'section-id'
             group_id = 'nsgroup-id'
             target_id = 'dummy'
-            self.nsxlib.firewall_section.create_rules(
-                None, section_id, group_id, False,
+            self.nsxlib.firewall_section.create_section_rules(
+                section_id, group_id, False,
                 "ALLOW", rules, {rule_id: target_id})
             add_rules.assert_called_once_with([
                 {'display_name': mock.ANY,
@@ -190,8 +213,8 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
                     section_id = 'section-id'
                     group_id = 'nsgroup-id'
                     target_id = 'dummy'
-                    self.nsxlib.firewall_section.create_rules(
-                        None, section_id, group_id, False,
+                    self.nsxlib.firewall_section.create_section_rules(
+                        section_id, group_id, False,
                         "ALLOW", rules, {rule_id: target_id})
                     add_rules.assert_called_once_with([
                         {'display_name': mock.ANY,
@@ -224,16 +247,18 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
         target_id = 'dummy'
         with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                         return_value="2.3.0"):
-            self.assertRaises(nsxlib_exc.InvalidInput,
-                              self.nsxlib.firewall_section.create_rules,
-                              None, section_id, group_id, False,
-                              "ALLOW", rules, {rule_id: target_id})
+            self.assertRaises(
+                nsxlib_exc.InvalidInput,
+                self.nsxlib.firewall_section.create_section_rules,
+                section_id, group_id, False,
+                "ALLOW", rules, {rule_id: target_id})
         with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                         return_value="2.4.0"):
-            self.assertRaises(nsxlib_exc.InvalidInput,
-                              self.nsxlib.firewall_section.create_rules,
-                              None, section_id, group_id, False,
-                              "ALLOW", rules, {rule_id: target_id})
+            self.assertRaises(
+                nsxlib_exc.InvalidInput,
+                self.nsxlib.firewall_section.create_section_rules,
+                section_id, group_id, False,
+                "ALLOW", rules, {rule_id: target_id})
 
     def test_create_rule_with_illegal_icmp_2_4(self):
         rule_id = uuidutils.generate_uuid()
@@ -250,10 +275,32 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
         target_id = 'dummy'
         with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                         return_value="2.4.0"):
-            self.assertRaises(nsxlib_exc.InvalidInput,
-                              self.nsxlib.firewall_section.create_rules,
-                              None, section_id, group_id, False,
-                              "ALLOW", rules, {rule_id: target_id})
+            self.assertRaises(
+                nsxlib_exc.InvalidInput,
+                self.nsxlib.firewall_section.create_section_rules,
+                section_id, group_id, False,
+                "ALLOW", rules, {rule_id: target_id})
+
+    def test_create_rule_with_illegal_icmp_0_255(self):
+        rule_id = uuidutils.generate_uuid()
+        rule = {'id': rule_id,
+                'ethertype': 'IPv4',
+                'protocol': 'icmp',
+                'direction': 'egress',
+                'port_range_min': 0,
+                'port_range_max': 255,
+                'remote_ip_prefix': None}
+        rules = [rule]
+        section_id = 'section-id'
+        group_id = 'nsgroup-id'
+        target_id = 'dummy'
+        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
+                        return_value="2.4.0"):
+            self.assertRaises(
+                nsxlib_exc.InvalidInput,
+                self.nsxlib.firewall_section.create_section_rules,
+                section_id, group_id, False,
+                "ALLOW", rules, {rule_id: target_id})
 
     def test_create_with_rules(self):
         expected_body = {
@@ -283,11 +330,47 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
                 '&action=create_with_rules'
             create.assert_called_with(resource, expected_body, headers=None)
 
+    def test_set_rule_logging(self):
+        section_id = '111'
+        rule_id = 1
+        orig_rule = {'id': rule_id, 'logged': False}
+        with mock.patch.object(self.nsxlib.firewall_section, 'get_rules',
+                               return_value={'results': [orig_rule]}),\
+            mock.patch.object(self.nsxlib.client, 'get', return_value={}),\
+            mock.patch.object(self.nsxlib.client, 'create') as update:
+            self.nsxlib.firewall_section.set_rule_logging(section_id, True)
+            update.assert_called_once_with(
+                'firewall/sections/111?action=update_with_rules',
+                {'rules': [{'id': 1, 'logged': True}]}, headers=None)
+
     def test_get_excludelist(self):
         with mock.patch.object(self.nsxlib.client, 'list') as clist:
             self.nsxlib.firewall_section.get_excludelist()
             clist.assert_called_with('firewall/excludelist')
 
+    def test_add_to_excludelist(self):
+        target_id = '111'
+        target_type = const.NSGROUP
+        with mock.patch.object(self.nsxlib.client, 'create') as create:
+            self.nsxlib.firewall_section.add_member_to_fw_exclude_list(
+                target_id, target_type)
+            create.assert_called_once_with(
+                'firewall/excludelist?action=add_member',
+                {'target_id': target_id, 'target_type': target_type},
+                headers=None)
+
+    def test_del_from_excludelist(self):
+        target_id = '111'
+        target_type = const.NSGROUP
+        with mock.patch.object(self.nsxlib.client, 'create') as create:
+            self.nsxlib.firewall_section.remove_member_from_fw_exclude_list(
+                target_id, target_type)
+            create.assert_called_once_with(
+                'firewall/excludelist?action=remove_member&'
+                'object_id=%s' % target_id,
+                None,
+                headers=None)
+
     def test_update(self):
         fws_tags = [{"scope": "name", "tag": "new_name"}]
         with mock.patch.object(self.nsxlib.client, 'update') as update:
@@ -299,6 +382,44 @@ class TestNsxLibFirewallSection(nsxlib_testcase.NsxLibTestCase):
                 data = {'tags': fws_tags}
                 update.assert_called_with(resource, data, headers=None)
 
+    def test_create_rules_using_add_rules(self):
+        revision = 5
+        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
+                        return_value='2.5.0'),\
+            mock.patch.object(self.nsxlib.client, 'get',
+                              return_value={'_revision': revision}),\
+            mock.patch.object(self.nsxlib.client, 'create') as create:
+            rule_id = uuidutils.generate_uuid()
+            rule = {'id': rule_id,
+                    'ethertype': 'IPv4',
+                    'protocol': 'ipip',
+                    'direction': 'ingress',
+                    'remote_ip_prefix': None}
+            rules = [rule]
+            section_id = 'section-id'
+            group_id = 'nsgroup-id'
+            target_id = 'dummy'
+            self.nsxlib.firewall_section.create_section_rules(
+                section_id, group_id, False,
+                "ALLOW", rules, {rule_id: target_id})
+            expected_rule = {'display_name': mock.ANY,
+                             'ip_protocol': 'IPV4',
+                             'direction': 'IN',
+                             'services': [{'service': {
+                                 'resource_type': 'IPProtocolNSService',
+                                 'protocol_number': 4}}],
+                             '_revision': revision,
+                             'disabled': False,
+                             'sources': [{'target_id': target_id,
+                                          'target_type': 'NSGroup'}],
+                             'destinations': [{'target_id': group_id,
+                                               'target_type': 'NSGroup'}],
+                             'logged': False, 'action': 'ALLOW'}
+            create.assert_called_once_with(
+                'firewall/sections/%s/rules?action=create_multiple&'
+                'operation=insert_bottom' % section_id,
+                {'rules': [expected_rule]})
+
 
 class TestNsxLibIPSet(nsxlib_testcase.NsxClientTestCase):
     """Tests for vmware_nsxlib.v3.security.NsxLibIPSet"""
@@ -332,7 +453,8 @@ class TestNsxLibIPSet(nsxlib_testcase.NsxClientTestCase):
         with mock.patch.object(self.nsxlib.client, 'delete') as delete:
             fake_ip_set = test_constants.FAKE_IP_SET.copy()
             self.nsxlib.ip_set.delete(fake_ip_set['id'])
-            delete.assert_called_with('ip-sets/%s' % fake_ip_set['id'])
+            delete.assert_called_with('ip-sets/%s' % fake_ip_set['id'],
+                                      headers=None)
 
     def test_update_ip_set(self):
         fake_ip_set = test_constants.FAKE_IP_SET.copy()
@@ -396,7 +518,10 @@ class TestNsxLibIPSet(nsxlib_testcase.NsxClientTestCase):
 class TestNsxLibNSGroup(nsxlib_testcase.NsxClientTestCase):
     """Tests for vmware_nsxlib.v3.security.NsxLibNSGroup"""
 
-    def test_get_nsgroup_complex_expression(self):
+    def setUp(self, *args, **kwargs):
+        super(TestNsxLibNSGroup, self).setUp(with_mocks=False)
+
+    def test_get_nsgroup_lp_complex_expression(self):
         port_tags = {'app': 'foo', 'project': 'myproject'}
         port_exp = [self.nsxlib.ns_group.get_port_tag_expression(k, v)
                     for k, v in six.iteritems(port_tags)]
@@ -406,6 +531,16 @@ class TestNsxLibNSGroup(nsxlib_testcase.NsxClientTestCase):
                         'expressions': port_exp}
         self.assertEqual(expected_exp, complex_exp)
 
+    def test_get_nsgroup_ls_complex_expression(self):
+        switch_tags = {'app': 'foo', 'project': 'myproject'}
+        switch_exp = [self.nsxlib.ns_group.get_switch_tag_expression(k, v)
+                      for k, v in six.iteritems(switch_tags)]
+        complex_exp = self.nsxlib.ns_group.get_nsgroup_complex_expression(
+            expressions=switch_exp)
+        expected_exp = {'resource_type': const.NSGROUP_COMPLEX_EXP,
+                        'expressions': switch_exp}
+        self.assertEqual(expected_exp, complex_exp)
+
     def test_update(self):
         nsg_tags = [{"scope": "name", "tag": "new_name"}]
         membership_criteria = []
@@ -420,12 +555,72 @@ class TestNsxLibNSGroup(nsxlib_testcase.NsxClientTestCase):
                         'membership_criteria': membership_criteria}
                 update.assert_called_with(resource, data, headers=None)
 
+    def test_list(self):
+        ns_group = {'id': 'dummy'}
+        ret_list = {'results': [ns_group]}
+        with mock.patch.object(self.nsxlib.ns_group.client, 'list',
+                               return_value=ret_list) as list_mock:
+            result = self.nsxlib.ns_group.list()
+            list_mock.assert_called_with('ns-groups?populate_references=false')
+            self.assertEqual([ns_group], result)
+
+    def test_find(self):
+        name = 'test_name'
+        ns_group = {'id': 'dummy', 'display_name': name}
+        ret_list = {'results': [ns_group]}
+        with mock.patch.object(self.nsxlib.ns_group.client, 'list',
+                               return_value=ret_list):
+            result = self.nsxlib.ns_group.find_by_display_name(name)
+            self.assertEqual([ns_group], result)
+
+    def test_read(self):
+        ns_group_id = 'dummy'
+        with mock.patch.object(self.nsxlib.ns_group.client, 'get') as get_mock:
+            self.nsxlib.ns_group.read(ns_group_id)
+            get_mock.assert_called_with(
+                'ns-groups/%s?populate_references=true' % ns_group_id)
+
+    def test_create(self):
+        name = 'test_name'
+        description = 'dummy'
+        with mock.patch.object(self.nsxlib.ns_group.client,
+                               'create') as create_mock:
+            self.nsxlib.ns_group.create(name, description, None)
+            create_mock.assert_called_with(
+                'ns-groups',
+                {'display_name': name,
+                 'description': description,
+                 'tags': None,
+                 'members': []})
+
+    def test_create_with_membership(self):
+        name = 'test_name'
+        description = 'dummy'
+        membership = 'criteria'
+        with mock.patch.object(self.nsxlib.ns_group.client,
+                               'create') as create_mock:
+            self.nsxlib.ns_group.create(name, description, None,
+                                        membership_criteria=membership)
+            create_mock.assert_called_with(
+                'ns-groups',
+                {'display_name': name,
+                 'description': description,
+                 'tags': None,
+                 'members': [],
+                 'membership_criteria': [membership]})
+
+    def test_delete(self):
+        ns_group_id = 'dummy'
+        with mock.patch.object(self.nsxlib.ns_group.client,
+                               'delete') as del_mock:
+            self.nsxlib.ns_group.delete(ns_group_id)
+            del_mock.assert_called_with(
+                'ns-groups/%s?force=true' % ns_group_id, headers=None)
+
     def test_update_nsgroup_and_section(self):
-        security_group = {
-            'name': 'name',
-            'id': uuidutils.generate_uuid(),
-            'description': None,
-            'logging': False}
+        name = 'name'
+        description = 'description'
+        logging = False
         nsgroup_id = uuidutils.generate_uuid()
         section_id = uuidutils.generate_uuid()
         log_sg_allowed_traffic = True
@@ -433,14 +628,14 @@ class TestNsxLibNSGroup(nsxlib_testcase.NsxClientTestCase):
         with mock.patch.object(self.nsxlib.client, 'update') as update_mock,\
             mock.patch.object(self.nsxlib.client, 'get') as get_mock:
             self.nsxlib.ns_group.update_nsgroup_and_section(
-                security_group, nsgroup_id, section_id,
+                name, description, logging, nsgroup_id, section_id,
                 log_sg_allowed_traffic)
             # updating the nsgroup and the section
             self.assertEqual(2, update_mock.call_count)
             # getting the rules, and get before each update
             self.assertEqual(3, get_mock.call_count)
 
-    def test_update_lport_nsgroups(self):
+    def test_update_lport_nsgroups_with_mocks(self):
         nsgroup_id1 = uuidutils.generate_uuid()
         nsgroup_id2 = uuidutils.generate_uuid()
         lport_id = uuidutils.generate_uuid()
@@ -457,3 +652,15 @@ class TestNsxLibNSGroup(nsxlib_testcase.NsxClientTestCase):
                                              [lport_id])
             remove_mock.assert_called_once_with(nsgroup_id1, 'LogicalPort',
                                                 lport_id)
+
+    def test_update_lport_nsgroups(self):
+        nsgroup_id1 = uuidutils.generate_uuid()
+        nsgroup_id2 = uuidutils.generate_uuid()
+        lport_id = uuidutils.generate_uuid()
+        original_nsgroups = [nsgroup_id1]
+        updated_nsgroups = [nsgroup_id2]
+
+        with mock.patch.object(self.nsxlib.client, 'create') as update_mock:
+            self.nsxlib.ns_group.update_lport_nsgroups(
+                lport_id, original_nsgroups, updated_nsgroups)
+            self.assertEqual(2, update_mock.call_count)
