@@ -29,10 +29,49 @@ LOG = log.getLogger(__name__)
 NULL_CURSOR_PREFIX = '0000'
 
 
+def get_http_error_details(response):
+    msg = response.json() if response.content else ''
+    error_code = None
+    related_error_codes = []
+
+    if isinstance(msg, dict) and 'error_message' in msg:
+        error_code = msg.get('error_code')
+        related_errors = [error['error_message'] for error in
+                          msg.get('related_errors', [])]
+        related_error_codes = [str(error['error_code']) for error in
+                               msg.get('related_errors', []) if
+                               error.get('error_code')]
+        msg = msg['error_message']
+        if related_errors:
+            msg += " relatedErrors: %s" % ' '.join(related_errors)
+
+    return {'status_code': response.status_code,
+            'error_code': error_code,
+            'related_error_codes': related_error_codes,
+            'details': msg}
+
+
+def init_http_exception_from_response(response):
+    if response is None or response:
+        # The response object has a __bool__ method that return True for
+        # status code under 400. In that case there is no need for exception
+        return None
+
+    error_details = get_http_error_details(response)
+    if not error_details['error_code']:
+        return None
+
+    error = http_error_to_exception(error_details['status_code'],
+                                    error_details['error_code'])
+
+    return error(manager='', **error_details)
+
+
 def http_error_to_exception(status_code, error_code):
     errors = {
         requests.codes.NOT_FOUND:
             {'202': exceptions.BackendResourceNotFound,
+             '500090': exceptions.StaleRevision,
              'default': exceptions.ResourceNotFound},
         requests.codes.BAD_REQUEST:
             {'60508': exceptions.NsxIndexingInProgress,
@@ -41,8 +80,10 @@ def http_error_to_exception(status_code, error_code):
              '8327': exceptions.NsxOverlapVlan,
              '500045': exceptions.NsxPendingDelete,
              '500030': exceptions.ResourceInUse,
+             '500087': exceptions.StaleRevision,
              '500105': exceptions.NsxOverlapAddresses,
-             '503040': exceptions.NsxSegemntWithVM},
+             '503040': exceptions.NsxSegemntWithVM,
+             '100148': exceptions.StaleRevision},
         requests.codes.CONFLICT: exceptions.StaleRevision,
         requests.codes.PRECONDITION_FAILED: exceptions.StaleRevision,
         requests.codes.INTERNAL_SERVER_ERROR:
@@ -50,7 +91,9 @@ def http_error_to_exception(status_code, error_code):
              '99': exceptions.ClientCertificateNotTrusted,
              '607': exceptions.APITransactionAborted},
         requests.codes.FORBIDDEN:
-            {'98': exceptions.BadXSRFToken},
+            {'98': exceptions.BadXSRFToken,
+             '403': exceptions.InvalidCredentials,
+             '505': exceptions.InvalidLicense},
         requests.codes.TOO_MANY_REQUESTS: exceptions.TooManyRequests,
         requests.codes.SERVICE_UNAVAILABLE: exceptions.ServiceUnavailable}
 
@@ -97,7 +140,7 @@ class RESTClient(object):
     def list(self, resource='', headers=None, silent=False):
         return self.url_list(resource, headers=headers, silent=silent)
 
-    def get(self, uuid, headers=None, silent=False, with_retries=True):
+    def get(self, uuid, headers=None, silent=False, with_retries=False):
         return self.url_get(uuid, headers=headers, silent=silent,
                             with_retries=with_retries)
 
@@ -130,7 +173,7 @@ class RESTClient(object):
             cursor = page.get('cursor', NULL_CURSOR_PREFIX)
         return concatenate_response
 
-    def url_get(self, url, headers=None, silent=False, with_retries=True):
+    def url_get(self, url, headers=None, silent=False, with_retries=False):
         return self._rest_call(url, method='GET', headers=headers,
                                silent=silent, with_retries=with_retries)
 
@@ -149,10 +192,10 @@ class RESTClient(object):
     def url_patch(self, url, body, headers=None):
         return self._rest_call(url, method='PATCH', body=body, headers=headers)
 
-    def _raise_error(self, status_code, operation, result_msg,
+    def _raise_error(self, operation, status_code, details,
                      error_code=None, related_error_codes=None):
         error = http_error_to_exception(status_code, error_code)
-        raise error(manager='', operation=operation, details=result_msg,
+        raise error(manager='', operation=operation, details=details,
                     error_code=error_code,
                     related_error_codes=related_error_codes,
                     status_code=status_code)
@@ -169,22 +212,9 @@ class RESTClient(object):
                                                    for code in expected]),
                              'body': result_msg})
 
-            error_code = None
-            related_error_codes = []
-            if isinstance(result_msg, dict) and 'error_message' in result_msg:
-                error_code = result_msg.get('error_code')
-                related_errors = [error['error_message'] for error in
-                                  result_msg.get('related_errors', [])]
-                related_error_codes = [str(error['error_code']) for error in
-                                       result_msg.get('related_errors', []) if
-                                       error.get('error_code')]
-                result_msg = result_msg['error_message']
-                if related_errors:
-                    result_msg += " relatedErrors: %s" % ' '.join(
-                        related_errors)
-            self._raise_error(result.status_code, operation, result_msg,
-                              error_code=error_code,
-                              related_error_codes=related_error_codes)
+            error_details = get_http_error_details(result)
+
+            self._raise_error(operation, **error_details)
 
     @classmethod
     def merge_headers(cls, *headers):
@@ -215,30 +245,30 @@ class RESTClient(object):
                    silent=False, expected_results=None, **kwargs):
         request_headers = headers.copy() if headers else {}
         request_headers.update(self._default_headers)
+
         if utils.INJECT_HEADERS_CALLBACK:
             inject_headers = utils.INJECT_HEADERS_CALLBACK()
             request_headers.update(inject_headers)
 
         request_url = self._build_url(url)
-
         do_request = getattr(self._conn, method.lower())
-        if not silent:
-
-            LOG.debug("REST call: %s %s. Headers: %s. Body: %s",
-                      method, request_url,
-                      utils.censor_headers(request_headers),
-                      self._mask_password(body))
-
+        if silent:
+            self._conn.set_silent(True)
         ts = time.time()
         result = do_request(
             request_url,
             data=body,
             headers=request_headers)
         te = time.time()
+        if silent:
+            self._conn.set_silent(False)
 
         if not silent:
-            LOG.debug("REST call: %s %s. Response: %s. Took %2.4f",
+            LOG.debug("REST call: %s %s. Headers: %s. Body: %s. Response: %s. "
+                      "Took %2.4f",
                       method, request_url,
+                      utils.censor_headers(request_headers),
+                      self._mask_password(body),
                       result.json() if result.content else '',
                       te - ts)
 
@@ -281,6 +311,12 @@ class NSX3Client(JSONRESTClient):
     NSX_V1_API_PREFIX = 'api/v1/'
     NSX_POLICY_V1_API_PREFIX = 'policy/api/v1/'
 
+    # NOTE: For user-facing client, NsxClusteredAPI instance
+    # will be passed as connection parameter below, thus all
+    # requests on this client will pass via cluster code to
+    # determine endpoint
+    # For validation client, TimeoutSession with specific
+    # endpoint parameters will be passed as connection.
     def __init__(self, connection, url_prefix=None,
                  default_headers=None,
                  nsx_api_managers=None,
@@ -313,35 +349,21 @@ class NSX3Client(JSONRESTClient):
             default_headers=default_headers,
             client_obj=client_obj)
 
-    def _raise_error(self, status_code, operation, result_msg,
+    def _raise_error(self, operation, status_code, details,
                      error_code=None, related_error_codes=None):
         """Override the Rest client errors to add the manager IPs"""
         error = http_error_to_exception(status_code, error_code)
         raise error(manager=self.nsx_api_managers,
                     operation=operation,
-                    details=result_msg,
+                    details=details,
                     error_code=error_code,
                     related_error_codes=related_error_codes,
                     status_code=status_code)
 
     def _rest_call(self, url, **kwargs):
-        if kwargs.get('with_retries', True):
-            # Retry on "607: Persistence layer is currently reconfiguring"
-            # and on "98: Cannot connect to server"
-            retry_codes = [exceptions.APITransactionAborted,
-                           exceptions.CannotConnectToServer]
-            if self.rate_limit_retry:
-                # If too many requests are handled by the nsx at the same time,
-                # error "429: Too Many Requests" or "503: Server Unavailable"
-                # will be returned.
-                retry_codes.append(exceptions.ServerBusy)
+        if 'with_retries' in kwargs and kwargs['with_retries']:
+            LOG.warning("with_retries setting is deprecated and will be "
+                        "removed. Please use exceptions setting in nsxlib "
+                        "config instead")
 
-            # the client is expected to retry after a random 400-600 milli,
-            # and later exponentially until 5 seconds wait
-            @utils.retry_random_upon_exception(
-                tuple(retry_codes), max_attempts=self.max_attempts)
-            def _rest_call_with_retry(self, url, **kwargs):
-                return super(NSX3Client, self)._rest_call(url, **kwargs)
-            return _rest_call_with_retry(self, url, **kwargs)
-        else:
-            return super(NSX3Client, self)._rest_call(url, **kwargs)
+        return super(NSX3Client, self)._rest_call(url, **kwargs)
