@@ -13,6 +13,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 #
+
 import abc
 import contextlib
 import copy
@@ -21,6 +22,7 @@ import inspect
 import itertools
 import logging
 import re
+import time
 
 import eventlet
 from eventlet import greenpool
@@ -30,14 +32,13 @@ from oslo_log import log
 from oslo_service import loopingcall
 import requests
 from requests import adapters
-from requests import exceptions as requests_exceptions
 import six
 import six.moves.urllib.parse as urlparse
-import tenacity
 import urllib3
 
 from vmware_nsxlib._i18n import _
 from vmware_nsxlib.v3 import client as nsx_client
+from vmware_nsxlib.v3 import constants
 from vmware_nsxlib.v3 import exceptions
 from vmware_nsxlib.v3 import utils
 
@@ -68,11 +69,6 @@ class AbstractHTTPProvider(object):
         """A unique string name for this provider."""
         pass
 
-    @abc.abstractmethod
-    def validate_connection(self, cluster_api, endpoint, conn):
-        """Validate the said connection for the given endpoint and cluster."""
-        pass
-
     @abc.abstractmethod
     def new_connection(self, cluster_api, provider):
         """Create a new http connection.
@@ -83,20 +79,6 @@ class AbstractHTTPProvider(object):
         """
         pass
 
-    @abc.abstractmethod
-    def is_connection_exception(self, exception):
-        """Determine if the given exception is related to connection failure.
-
-        Return True if it's a connection exception and False otherwise.
-        """
-
-    @abc.abstractmethod
-    def is_timeout_exception(self, exception):
-        """Determine if the given exception is related to timeout.
-
-        Return True if it's a timeout exception and False otherwise.
-        """
-
 
 class TimeoutSession(requests.Session):
     """Extends requests.Session to support timeout at the session level."""
@@ -105,6 +87,7 @@ class TimeoutSession(requests.Session):
         self.timeout = timeout
         self.read_timeout = read_timeout
         self.cert_provider = None
+        self._silent = False
         super(TimeoutSession, self).__init__()
 
     @property
@@ -115,6 +98,9 @@ class TimeoutSession(requests.Session):
     def cert_provider(self, value):
         self._cert_provider = value
 
+    def set_silent(self, silent_mode):
+        self._silent = silent_mode
+
     # wrapper timeouts at the session level
     # see: https://goo.gl/xNk7aM
     def request(self, *args, **kwargs):
@@ -186,23 +172,21 @@ class NSXRequestsHTTPProvider(AbstractHTTPProvider):
         return "%s-%s" % (requests.__title__, requests.__version__)
 
     def validate_connection(self, cluster_api, endpoint, conn):
+        # Retry during validation can cause retry storm, thus limit
+        # max_attempts to 1
+        # on connection level, validation will be retried according to
+        # nsxlib 'retries' and 'http_timeout' parameters.
         client = nsx_client.NSX3Client(
             conn, url_prefix=endpoint.provider.url,
-            url_path_base=cluster_api.nsxlib_config.url_base,
-            default_headers=conn.default_headers)
+            url_path_base=nsx_client.NSX3Client.NSX_V1_API_PREFIX,
+            default_headers=conn.default_headers,
+            max_attempts=1)
 
-        # Check the manager state directly
-        if cluster_api.nsxlib_config.validate_connection_method:
-            cluster_api.nsxlib_config.validate_connection_method(
-                client, endpoint.provider.url)
-
-        # If keeplive section returns a list, it is assumed to be non-empty
-        keepalive_section = cluster_api.nsxlib_config.keepalive_section
-        result = client.get(keepalive_section, silent=True)
-        if not result or result.get('result_count', 1) <= 0:
-            msg = _("No %(section)s found "
-                    "for '%(url)s'") % {'section': keepalive_section,
-                                        'url': endpoint.provider.url}
+        # Try to get the status silently and with no retries
+        status = client.get('reverse-proxy/node/health',
+                            silent=True, with_retries=False)
+        if not status or not status.get('healthy', False):
+            msg = _("NSX Node is not healthy, reported status: %s") % status
             LOG.warning(msg)
             raise exceptions.ResourceNotFound(
                 manager=endpoint.provider.url, operation=msg)
@@ -249,18 +233,8 @@ class NSXRequestsHTTPProvider(AbstractHTTPProvider):
         self.get_default_headers(session, provider,
                                  config.allow_overwrite_header,
                                  config.token_provider)
-
         return session
 
-    def is_connection_exception(self, exception):
-        return isinstance(exception, requests_exceptions.ConnectionError)
-
-    def is_timeout_exception(self, exception):
-        return isinstance(exception, requests_exceptions.Timeout)
-
-    def is_conn_open_exception(self, exception):
-        return isinstance(exception, requests_exceptions.ConnectTimeout)
-
     def get_default_headers(self, session, provider, allow_overwrite_header,
                             token_provider=None):
         """Get the default headers that should be added to future requests"""
@@ -307,13 +281,16 @@ class NSXRequestsHTTPProvider(AbstractHTTPProvider):
         else:
             if resp.status_code != 200 and resp.status_code != 201:
                 LOG.warning("Session create failed for endpoint %s with "
-                            "response %s",
-                            provider.url, resp.status_code)
+                            "response %s, error message: %s, "
+                            "local NSX time: %s",
+                            provider.url, resp.status_code,
+                            resp.json().get('error_message'),
+                            resp.headers['Date'])
                 # this may will later cause the endpoint to be Down
             else:
                 for header_name in resp.headers:
                     if self.SET_COOKIE_FIELD.lower() == header_name.lower():
-                        m = re.match('%s=.*?\;' % self.JSESSIONID,  # noqa
+                        m = re.match(r'%s=.*?\;' % self.JSESSIONID,  # noqa
                                      resp.headers[header_name])
                         if m:
                             session.default_headers[self.COOKIE_FIELD] = (
@@ -391,11 +368,18 @@ class Endpoint(object):
     to the underlying connections.
     """
 
-    def __init__(self, provider, pool):
+    def __init__(self, provider, pool, api_rate_limit=None,
+                 api_rate_mode=None, api_call_collector=None):
         self.provider = provider
         self.pool = pool
         self._state = EndpointState.INITIALIZED
         self._last_updated = datetime.datetime.now()
+        if api_rate_mode == constants.API_RATE_MODE_AIMD:
+            self.rate_limiter = utils.APIRateLimiterAIMD(
+                max_calls=api_rate_limit)
+        else:
+            self.rate_limiter = utils.APIRateLimiter(max_calls=api_rate_limit)
+        self.api_call_collector = api_call_collector
 
     def regenerate_pool(self):
         self.pool = pools.Pool(min_size=self.pool.min_size,
@@ -425,6 +409,15 @@ class Endpoint(object):
 
         return old_state
 
+    def add_api_record(self, record):
+        if self.api_call_collector:
+            self.api_call_collector.add_record(record)
+
+    def pop_all_api_records(self):
+        if self.api_call_collector:
+            return self.api_call_collector.pop_all_records()
+        return []
+
     def __str__(self):
         return "[%s] %s" % (self.state, self.provider)
 
@@ -435,9 +428,11 @@ class EndpointConnection(object):
     Which contains an endpoint and a connection for that endpoint.
     """
 
-    def __init__(self, endpoint, connection):
+    def __init__(self, endpoint, connection, conn_wait, rate_wait):
         self.endpoint = endpoint
         self.connection = connection
+        self.conn_wait = conn_wait
+        self.rate_wait = rate_wait
 
 
 class ClusteredAPI(object):
@@ -453,14 +448,21 @@ class ClusteredAPI(object):
                  http_provider,
                  min_conns_per_pool=0,
                  max_conns_per_pool=20,
-                 keepalive_interval=33):
+                 keepalive_interval=33,
+                 api_rate_limit=None,
+                 api_rate_mode=None,
+                 api_log_mode=None):
 
         self._http_provider = http_provider
         self._keepalive_interval = keepalive_interval
+        self._print_keepalive = 0
+        self._silent = False
+        self._api_call_collectors = []
 
         def _init_cluster(*args, **kwargs):
-            self._init_endpoints(providers,
-                                 min_conns_per_pool, max_conns_per_pool)
+            self._init_endpoints(providers, min_conns_per_pool,
+                                 max_conns_per_pool, api_rate_limit,
+                                 api_rate_mode, api_log_mode)
 
         _init_cluster()
 
@@ -469,8 +471,12 @@ class ClusteredAPI(object):
         # loops + state
         self._reinit_cluster = _init_cluster
 
-    def _init_endpoints(self, providers,
-                        min_conns_per_pool, max_conns_per_pool):
+    def set_silent(self, silent_mode):
+        self._silent = silent_mode
+
+    def _init_endpoints(self, providers, min_conns_per_pool,
+                        max_conns_per_pool, api_rate_limit, api_rate_mode,
+                        api_log_mode):
         LOG.debug("Initializing API endpoints")
 
         def _create_conn(p):
@@ -479,6 +485,14 @@ class ClusteredAPI(object):
 
             return _conn
 
+        self._api_call_collectors = []
+        api_call_collector = None
+        if api_log_mode == constants.API_CALL_LOG_PER_CLUSTER:
+            # Init one instance of collector for the entire cluster
+            api_call_collector = utils.APICallCollector(
+                ",".join([provider.id for provider in providers]))
+            self._api_call_collectors.append(api_call_collector)
+
         self._endpoints = {}
         for provider in providers:
             pool = pools.Pool(
@@ -487,7 +501,13 @@ class ClusteredAPI(object):
                 order_as_stack=True,
                 create=_create_conn(provider))
 
-            endpoint = Endpoint(provider, pool)
+            if api_log_mode == constants.API_CALL_LOG_PER_ENDPOINT:
+                # Init one instance of collector for each endpoint
+                api_call_collector = utils.APICallCollector(provider.id)
+                self._api_call_collectors.append(api_call_collector)
+
+            endpoint = Endpoint(provider, pool, api_rate_limit, api_rate_mode,
+                                api_call_collector)
             self._endpoints[provider.id] = endpoint
 
         # service requests using round robin
@@ -508,13 +528,18 @@ class ClusteredAPI(object):
                 break
             eventlet.sleep(0.5)
 
-        for endpoint in self._endpoints.values():
-            # dynamic loop for each endpoint to ensure connectivity
-            loop = loopingcall.DynamicLoopingCall(
-                self._endpoint_keepalive, endpoint)
-            loop.start(initial_delay=self._keepalive_interval,
-                       periodic_interval_max=self._keepalive_interval,
-                       stop_on_exception=False)
+        if len(self._endpoints) > 1:
+            # We don't monitor connectivity when one endpoint is available,
+            # since there is no alternative to querying this single backend
+            # If endpoint was down, we can tolerate extra roundtrip to
+            # validate connectivity
+            for endpoint in self._endpoints.values():
+                # dynamic loop for each endpoint to ensure connectivity
+                loop = loopingcall.DynamicLoopingCall(
+                    self._endpoint_keepalive, endpoint)
+                loop.start(initial_delay=self._keepalive_interval,
+                           periodic_interval_max=self._keepalive_interval,
+                           stop_on_exception=False)
 
         LOG.debug("Done initializing API endpoint(s). "
                   "API cluster health: %s", self.health)
@@ -523,6 +548,13 @@ class ClusteredAPI(object):
         delta = datetime.datetime.now() - endpoint.last_updated
         if delta.seconds >= self._keepalive_interval:
             # TODO(boden): backoff on validation failure
+            if self._print_keepalive % 10 == 0:
+                # Print keepalive debug message once every 10 probes
+                LOG.debug("Running keepalive probe for cluster endpoint "
+                          "'%(ep)s' ",
+                          {'ep': endpoint})
+            self._print_keepalive += 1
+
             self._validate(endpoint)
             return self._keepalive_interval
         return self._keepalive_interval - delta.seconds
@@ -539,6 +571,10 @@ class ClusteredAPI(object):
     def http_provider(self):
         return self._http_provider
 
+    @property
+    def api_call_collectors(self):
+        return self._api_call_collectors
+
     @property
     def health(self):
         down = 0
@@ -558,22 +594,30 @@ class ClusteredAPI(object):
     def _validate(self, endpoint):
         try:
             with endpoint.pool.item() as conn:
-                self._http_provider.validate_connection(self, endpoint, conn)
+                # with some configurations, validation will be skipped
+                self._http_provider.validate_connection(self,
+                                                        endpoint,
+                                                        conn)
                 endpoint.set_state(EndpointState.UP)
-        except exceptions.ClientCertificateNotTrusted:
-            LOG.warning("Failed to validate API cluster endpoint "
-                        "'%(ep)s' due to untrusted client certificate",
-                        {'ep': endpoint})
-            # regenerate connection pool based on new certificate
-            endpoint.regenerate_pool()
-        except exceptions.BadXSRFToken:
-            LOG.warning("Failed to validate API cluster endpoint "
-                        "'%(ep)s' due to expired XSRF token",
-                        {'ep': endpoint})
-            # regenerate connection pool based on token
-            endpoint.regenerate_pool()
         except Exception as e:
-            endpoint.set_state(EndpointState.DOWN)
+            if self.nsxlib_config.exception_config.should_regenerate(e):
+                LOG.warning("Failed to validate API cluster endpoint "
+                            "'%(ep)s' due to an exception that calls for "
+                            "regeneration. Re-generating pool.",
+                            {'ep': endpoint})
+                if bool(self.nsxlib_config.token_provider):
+                    # get new jwt token for authentication
+                    self.nsxlib_config.token_provider.get_token(
+                        refresh_token=True)
+                # refresh endpoint with new headers that have updated token
+                endpoint.regenerate_pool()
+                return
+            elif self.nsxlib_config.exception_config.should_retry(e):
+                LOG.info("Exception is retriable, endpoint stays UP")
+                endpoint.set_state(EndpointState.UP)
+            else:
+                endpoint.set_state(EndpointState.DOWN)
+
             LOG.warning("Failed to validate API cluster endpoint "
                         "'%(ep)s' due to: %(err)s",
                         {'ep': endpoint, 'err': e})
@@ -596,22 +640,13 @@ class ClusteredAPI(object):
                     return endpoint
                 seen += 1
 
-        @utils.retry_upon_none_result(self.nsxlib_config.max_attempts)
-        def _select_endpoint_internal_with_retry():
-            # redo endpoint selection with refreshing states
-            return _select_endpoint_internal(refresh=True)
-
         # First attempt to get an UP endpoint
         endpoint = _select_endpoint_internal()
         if endpoint or not self.nsxlib_config.cluster_unavailable_retry:
             return endpoint
 
         # Retry the selection while refreshing the endpoints state
-        try:
-            return _select_endpoint_internal_with_retry()
-        except tenacity.RetryError:
-            # exhausted number of retries
-            return None
+        return _select_endpoint_internal(refresh=True)
 
     def endpoint_for_connection(self, conn):
         # check all endpoint pools
@@ -647,18 +682,68 @@ class ClusteredAPI(object):
                      {'ep': endpoint,
                       'max': endpoint.pool.max_size,
                       'waiting': endpoint.pool.waiting()})
+            conn_wait_start = time.time()
+        else:
+            conn_wait_start = None
         # pool.item() will wait if pool has 0 free
         with endpoint.pool.item() as conn:
-            yield EndpointConnection(endpoint, conn)
+            if conn_wait_start:
+                conn_wait = time.time() - conn_wait_start
+            else:
+                conn_wait = 0
+            with endpoint.rate_limiter as rate_wait:
+                # Connection validation calls are not currently rate-limited
+                # by this context manager.
+                # This should be fine as validation api calls are sent in a
+                # slow rate at once per 33 seconds by default.
+                yield EndpointConnection(endpoint, conn, conn_wait, rate_wait)
+
+    def _raise_http_exception_if_needed(self, response, endpoint):
+        # We need to inspect http codes to understand whether
+        # this error is relevant for endpoint-level decisions, such
+        # as ground endpoint or retry with next endpoint
+        exc = nsx_client.init_http_exception_from_response(response)
+        if not exc:
+            # This exception is irrelevant for endpoint decisions
+            return
+
+        if self.nsxlib_config.exception_config.should_regenerate(exc):
+            if bool(self.nsxlib_config.token_provider):
+                # get new jwt token for authentication
+                self.nsxlib_config.token_provider.get_token(
+                    refresh_token=True)
+            # refresh endpoint so that it gets new header with updated token
+            endpoint.regenerate_pool()
+            raise exc
+
+        exc_config = self.nsxlib_config.exception_config
+        if (exc_config.should_ground_endpoint(exc) or
+            exc_config.should_retry(exc)):
+            raise exc
 
     def _proxy_stub(self, proxy_for):
         def _call_proxy(url, *args, **kwargs):
-            return self._proxy(proxy_for, url, *args, **kwargs)
+            try:
+                return self._proxy(proxy_for, url, *args, **kwargs)
+            except Exception as ex:
+                # If this was exception that grounded the cluster,
+                # we want to translate this exception to
+                # ServiceClusterUnavailable. This is in order to
+                # provide unified "cluster down" experience for
+                # the client
+                exc_config = self.nsxlib_config.exception_config
+                if exc_config.should_ground_endpoint(ex):
+                    raise exceptions.ServiceClusterUnavailable(
+                        cluster_id=self.cluster_id)
+
+                raise ex
+
         return _call_proxy
 
     def _proxy(self, proxy_for, uri, *args, **kwargs):
 
-        @utils.retry_upon_none_result(self.nsxlib_config.max_attempts)
+        @utils.retry_random_upon_exception_result(
+            max_attempts=self.nsxlib_config.max_attempts)
         def _proxy_internal(proxy_for, uri, *args, **kwargs):
             # proxy http request call to an avail endpoint
             with self.endpoint_connection() as conn_data:
@@ -672,34 +757,59 @@ class ClusteredAPI(object):
                     uri = "/%s" % uri
                 url = "%s%s" % (endpoint.provider.url, uri)
                 try:
-                    LOG.debug("API cluster proxy %s %s to %s",
-                              proxy_for.upper(), uri, url)
                     # Add the connection default headers
                     if conn.default_headers:
                         kwargs['headers'] = kwargs.get('headers', {})
                         kwargs['headers'].update(conn.default_headers)
+                    if not self._silent:
+                        # To censor sensitive headers before logging
+                        kwargs_copy = copy.copy(kwargs)
+                        kwargs_copy['headers'] = utils.censor_headers(
+                            kwargs_copy['headers'])
+                        LOG.debug("API cluster proxy %s %s to %s with %s. "
+                                  "Waited conn: %2.4f, rate: %2.4f",
+                                  proxy_for.upper(), uri, url, kwargs_copy,
+                                  conn_data.conn_wait, conn_data.rate_wait)
 
                     # call the actual connection method to do the
                     # http request/response over the wire
                     response = do_request(url, *args, **kwargs)
                     endpoint.set_state(EndpointState.UP)
 
+                    # add api call log
+                    api_record = utils.APICallRecord(
+                        verb=proxy_for, uri=uri, status=response.status_code,
+                        provider=endpoint.provider.id)
+                    endpoint.add_api_record(api_record)
+
+                    # Adjust API Rate Limit before raising HTTP exception
+                    endpoint.rate_limiter.adjust_rate(
+                        wait_time=conn_data.rate_wait,
+                        status_code=response.status_code)
+
+                    # for some status codes, we need to bring the cluster
+                    # down or retry API call
+                    self._raise_http_exception_if_needed(response, endpoint)
+
                     return response
                 except Exception as e:
                     LOG.warning("Request failed due to: %s", e)
-                    if (not self._http_provider.is_connection_exception(e) and
-                        not self._http_provider.is_timeout_exception(e)):
-                        # only trap and retry connection & timeout errors
-                        raise e
-                    if self._http_provider.is_conn_open_exception(e):
-                        # unable to establish new connection - endpoint is
-                        # inaccessible
+                    exc_config = self.nsxlib_config.exception_config
+                    if exc_config.should_ground_endpoint(e):
+                        # consider endpoint inaccessible and move to next
+                        # endpoint
                         endpoint.set_state(EndpointState.DOWN)
 
-                    LOG.info("Connection to %s failed, checking additional "
-                             "connections and endpoints" % url)
-                    # this might be a result of server closing connection
-                    # return None so it will retry upto max_attempts.
+                    elif not exc_config.should_retry(e):
+                        LOG.info("Exception %s is configured as not retriable",
+                                 e)
+                        raise e
+
+                    # Returning the exception instead of raising it will cause
+                    # decorator to retry. If retry attempts is exceeded, this
+                    # same exception will be raised due to overriden reraise
+                    # method of RetryAttemptsExceeded
+                    return e
 
         return _proxy_internal(proxy_for, uri, *args, **kwargs)
 
@@ -717,7 +827,10 @@ class NSXClusteredAPI(ClusteredAPI):
             self._build_conf_providers(),
             self._http_provider,
             max_conns_per_pool=self.nsxlib_config.concurrent_connections,
-            keepalive_interval=self.nsxlib_config.conn_idle_timeout)
+            keepalive_interval=self.nsxlib_config.conn_idle_timeout,
+            api_rate_limit=self.nsxlib_config.api_rate_limit_per_endpoint,
+            api_rate_mode=self.nsxlib_config.api_rate_mode,
+            api_log_mode=self.nsxlib_config.api_log_mode)
 
         LOG.debug("Created NSX clustered API with '%s' "
                   "provider", self._http_provider.provider_id)
