@@ -71,7 +71,7 @@ class NsxLibPortMirror(utils.NsxLibApiBase):
         :param mirror_session_id: string representing the UUID of the port
                                   mirror session to be deleted.
         """
-        self.client.delete(self.get_path(mirror_session_id))
+        self._delete_with_retry(mirror_session_id)
 
 
 class NsxLibBridgeEndpointProfile(utils.NsxLibApiBase):
@@ -126,7 +126,7 @@ class NsxLibBridgeEndpointProfile(utils.NsxLibApiBase):
                                            the bridge endpoint profile to be
                                            deleted.
         """
-        self.client.delete(self.get_path(bridge_endpoint_profile_id))
+        self._delete_with_retry(bridge_endpoint_profile_id)
 
 
 class NsxLibBridgeEndpoint(utils.NsxLibApiBase):
@@ -165,7 +165,7 @@ class NsxLibBridgeEndpoint(utils.NsxLibApiBase):
         :param bridge_endpoint_id: string representing the UUID of the bridge
                                    endpoint to be deleted.
         """
-        self.client.delete(self.get_path(bridge_endpoint_id))
+        self._delete_with_retry(bridge_endpoint_id)
 
 
 class NsxLibLogicalSwitch(utils.NsxLibApiBase):
@@ -556,7 +556,7 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
             if utils.dict_match(kwargs, res):
                 LOG.debug("Deleting %s from resource %s", res, resource)
                 delete_resource = resource + "/" + str(res['id'])
-                self.client.delete(delete_resource)
+                self._delete_by_path_with_retry(delete_resource)
                 matched_num = matched_num + 1
         if matched_num == 0:
             if skip_not_found:
@@ -650,9 +650,9 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
         return self.client.create(resource, body)
 
     def delete_static_route(self, logical_router_id, static_route_id):
-        resource = 'logical-routers/%s/routing/static-routes/%s' % (
+        path = 'logical-routers/%s/routing/static-routes/%s' % (
             logical_router_id, static_route_id)
-        self.client.delete(resource)
+        self._delete_by_path_with_retry(path)
 
     def delete_static_route_by_values(self, logical_router_id,
                                       dest_cidr=None, nexthop=None):
@@ -665,10 +665,15 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
             kwargs['next_hops'] = [{"ip_address": nexthop}]
         return self._delete_resource_by_values(resource, **kwargs)
 
+    def list_static_routes(self, logical_router_id):
+        resource = ('logical-routers/%s/routing/static-routes' %
+                    logical_router_id)
+        return self.client.list(resource)
+
     def delete_nat_rule(self, logical_router_id, nat_rule_id):
-        resource = 'logical-routers/%s/nat/rules/%s' % (logical_router_id,
-                                                        nat_rule_id)
-        self.client.delete(resource)
+        path = 'logical-routers/%s/nat/rules/%s' % (logical_router_id,
+                                                    nat_rule_id)
+        self._delete_by_path_with_retry(path)
 
     def delete_nat_rule_by_values(self, logical_router_id,
                                   strict_mode=True,
@@ -801,7 +806,7 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
         url = lrouter_id
         if force:
             url += '?force=%s' % force
-        return self.client.delete(self.get_path(url))
+        return self._delete_by_path_with_retry(self.get_path(url))
 
     def update(self, lrouter_id, *args, **kwargs):
         body = {}
@@ -826,6 +831,7 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
 
         If the router was already retrieved from the backend it is possible
         to give it as an input to avoid another backend call.
+        In case of multiple sections, the first valid one will be returned
         """
         if not router_body:
             router_body = self.get(lrouter_id)
@@ -834,7 +840,7 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
             for sec in firewall_sections:
                 if (sec.get('is_valid') and
                     sec.get('target_type') == "FirewallSection"):
-                    return firewall_sections[0].get('target_id')
+                    return sec.get('target_id')
 
     def list(self, router_type=None):
         """List all/by type logical routers."""
@@ -863,6 +869,11 @@ class NsxLibLogicalRouter(utils.NsxLibApiBase):
         resource = ('logical-routers/%s/routing/bgp' % logical_router_id)
         return self.client.get(resource)
 
+    def get_bgp_neighbors(self, logical_router_id):
+        resource = ('logical-routers/%s/routing/bgp/neighbors' %
+                    logical_router_id)
+        return self.client.get(resource)
+
     def get_route_map(self, logical_router_id, route_map_id):
         resource = ('logical-routers/%s/routing/route-maps/%s' % (
                     logical_router_id, route_map_id))
@@ -948,9 +959,11 @@ class NsxLibTransportNode(utils.NsxLibApiBase):
                 return []
 
             host_switches = tn.get('host_switch_spec').get('host_switches', [])
-
-            return [ep.get('transport_zone_id') for ep in
-                    host_switches[0].get('transport_zone_endpoints', [])]
+            tzs = []
+            for host_switch in host_switches:
+                tzs.extend([ep.get('transport_zone_id') for ep in
+                            host_switch.get('transport_zone_endpoints', [])])
+            return tzs
         else:
             return [ep.get('transport_zone_id') for ep in
                     tn.get('transport_zone_endpoints', [])]
@@ -1073,9 +1086,13 @@ class NsxLibIpBlockSubnet(utils.NsxLibApiBase):
             headers = {'X-Allow-Overwrite': 'true'}
         return self.client.create(self.get_path(), body, headers=headers)
 
-    def delete(self, subnet_id):
+    def delete(self, subnet_id, allow_overwrite=False):
         """Delete a IP block subnet on the backend."""
-        self.client.delete(self.get_path(subnet_id))
+        headers = None
+        if allow_overwrite:
+            # Need to force delete subnet if IpBlock is created by Policy API
+            headers = {'X-Allow-Overwrite': 'true'}
+        self._delete_with_retry(subnet_id, headers=headers)
 
     def list(self, ip_block_id):
         resource = '%s?block_id=%s' % (self.get_path(), ip_block_id)
