@@ -151,6 +151,14 @@ class BadXSRFToken(ManagerError):
     message = _("Bad or expired XSRF token")
 
 
+class InvalidCredentials(ManagerError):
+    message = _("Failed to authenticate with NSX: %(msg)s")
+
+
+class InvalidLicense(ManagerError):
+    message = _("No valid License to configure NSX resources: %(msg)s")
+
+
 class BadJSONWebTokenProviderRequest(NsxLibException):
     message = _("Bad or expired JSON web token request from provider: %(msg)s")
 
@@ -170,11 +178,6 @@ class NSGroupIsFull(ManagerError):
                 "capacity, unable to add additional members.")
 
 
-class NumberOfNsgroupCriteriaTagsReached(ManagerError):
-    message = _("Port can be associated with at most %(max_num)s "
-                "security-groups.")
-
-
 class SecurityGroupMaximumCapacityReached(ManagerError):
     message = _("Security Group %(sg_id)s has reached its maximum capacity, "
                 "no more ports can be associated with this security-group.")
