@@ -153,7 +153,7 @@ class LoadBalancerBase(utils.NsxLibApiBase):
 
     def delete(self, object_id):
         object_url = self.resource + '/' + object_id
-        return self.client.delete(object_url)
+        return self._delete_by_path_with_retry(object_url)
 
 
 class ApplicationProfile(LoadBalancerBase):
@@ -490,10 +490,10 @@ class Service(LoadBalancerBase):
                                       'virtual-servers', 'status')
         return self.client.get(object_url)
 
-    def get_stats(self, service_id, source='realtime'):
+    def get_stats(self, service_id, source='realtime', silent=False):
         object_url = '%s/%s/%s?source=%s' % (self.resource, service_id,
                                              'statistics', source)
-        return self.client.get(object_url)
+        return self.client.get(object_url, silent=silent)
 
     def get_usage(self, service_id):
         object_url = '%s/%s/%s' % (self.resource, service_id, 'usage')
