@@ -33,6 +33,10 @@ VIF_TYPE_CHILD = "CHILD"
 
 ALLOCATE_ADDRESS_NONE = "None"
 
+# SegmentPort init_state types
+INIT_STATE_UNBLOCKED_VLAN = 'UNBLOCKED_VLAN'
+INIT_STATE_RESTORE_VIF = 'RESTORE_VIF'
+
 # NSXv3 L2 Gateway constants
 BRIDGE_ENDPOINT = "BRIDGEENDPOINT"
 FAILOVER_MODE_PREEMPTIVE = "PREEMPTIVE"
@@ -110,10 +114,10 @@ IPV4_IPV6 = 'IPV4_IPV6'
 
 LOCAL_IP_PREFIX = 'local_ip_prefix'
 
-LOGGING = 'logging'
-
 # Allowed address pairs
 NUM_ALLOWED_IP_ADDRESSES = 128
+NUM_ALLOWED_IP_ADDRESSES_v4 = NUM_ALLOWED_IP_ADDRESSES
+NUM_ALLOWED_IP_ADDRESSES_v6 = 15
 MAX_STATIC_ROUTES = 26
 
 # QoS directions egress/ingress
@@ -147,6 +151,8 @@ NSX_VERSION_2_3_0 = '2.3.0'
 NSX_VERSION_2_4_0 = '2.4.0'
 NSX_VERSION_2_5_0 = '2.5.0'
 NSX_VERSION_3_0_0 = '3.0.0'
+NSX_VERSION_3_0_2 = '3.0.2'
+NSX_VERSION_3_1_0 = '3.1.0'
 
 # Features available depending on the NSX Manager backend version
 FEATURE_MAC_LEARNING = 'MAC Learning'
@@ -173,6 +179,10 @@ FEATURE_RELAX_SCALE_VALIDATION = 'Relax Scale Validation for LbService'
 FEATURE_SWITCH_HYPERBUS_MODE = 'Switch hyperbus mode with policy API'
 FEATURE_GET_TZ_FROM_SWITCH = 'Get TZ endpoints from host switch'
 FEATURE_ROUTE_REDISTRIBUTION_CONFIG = 'Tier0 route redistribution config'
+FEATURE_CONTAINER_CLUSTER_INVENTORY = 'Container Cluster Inventory'
+FEATURE_IPV6 = 'IPV6 Forwarding and Address Allocation'
+FEATURE_MP2P_MIGRATION = 'MP to Policy Migration'
+FEATURE_SPOOFGUARD_CIDR = 'Spoofguard IPv4 CIDR'
 
 # Features available depending on the Policy Manager backend version
 FEATURE_NSX_POLICY = 'NSX Policy'
@@ -181,8 +191,3 @@ FEATURE_NSX_POLICY_MDPROXY = 'NSX Policy Metadata Proxy'
 FEATURE_NSX_POLICY_DHCP = 'NSX Policy DHCP'
 FEATURE_NSX_POLICY_GLOBAL_CONFIG = 'NSX Policy Global Config'
 FEATURE_NSX_POLICY_ADMIN_STATE = 'NSX Policy Segment admin state'
-
-# FEATURE available depending on Inventory service backend version
-FEATURE_CONTAINER_CLUSTER_INVENTORY = 'Container Cluster Inventory'
-
-FEATURE_IPV6 = 'IPV6 Forwarding and Address Allocation'
