@@ -18,16 +18,16 @@ import abc
 from distutils import version
 
 from oslo_log import log as logging
+import requests
 import six
 
+from vmware_nsxlib.v3 import exceptions
 from vmware_nsxlib.v3 import nsx_constants
 from vmware_nsxlib.v3.policy import constants
 from vmware_nsxlib.v3 import utils
 
 LOG = logging.getLogger(__name__)
 
-LOG = logging.getLogger(__name__)
-
 TENANTS_PATH_PATTERN = "%s/"
 DOMAINS_PATH_PATTERN = TENANTS_PATH_PATTERN + "domains/"
 IP_BLOCKS_PATH_PATTERN = TENANTS_PATH_PATTERN + "ip-blocks/"
@@ -59,8 +59,14 @@ WAF_PROFILES_PATH_PATTERN = (TENANTS_PATH_PATTERN +
 CERTIFICATE_PATH_PATTERN = TENANTS_PATH_PATTERN + "certificates/"
 EXCLUDE_LIST_PATH_PATTERN = (TENANTS_PATH_PATTERN +
                              "settings/firewall/security/exclude-list")
-
-REALIZATION_PATH = "infra/realized-state/realized-entities?intent_path=%s"
+REALIZATION_PATH_PREFIX = "infra/realized-state/"
+REALIZATION_PATH_SUFFIX = "?intent_path=%s"
+REALIZATION_PATH = (REALIZATION_PATH_PREFIX + "realized-entities" +
+                    REALIZATION_PATH_SUFFIX)
+REALIZATION_REFRESH_PATH = (REALIZATION_PATH_PREFIX + "realized-entity" +
+                            REALIZATION_PATH_SUFFIX + "&action=refresh")
+REALIZATION_STATUS_PATH = (REALIZATION_PATH_PREFIX + "status" +
+                           REALIZATION_PATH_SUFFIX)
 DHCP_REALY_PATTERN = TENANTS_PATH_PATTERN + "dhcp-relay-configs/"
 DHCP_SERVER_PATTERN = TENANTS_PATH_PATTERN + "dhcp-server-configs/"
 MDPROXY_PATTERN = TENANTS_PATH_PATTERN + "metadata-proxies/"
@@ -785,27 +791,29 @@ class Subnet(object):
         self.dhcp_ranges = dhcp_ranges
         self.dhcp_config = dhcp_config
 
-    def get_obj_dict(self):
+    def get_obj_dict(self, nsx_version):
         body = {'gateway_address': self.gateway_address}
+
         if self.dhcp_ranges:
             body['dhcp_ranges'] = self.dhcp_ranges
-        if self.dhcp_config:
+
+        if self.dhcp_config is not None:
             body['dhcp_config'] = (
                 self.dhcp_config.get_obj_dict()
-                if isinstance(self.dhcp_config, SegmentDhcpConfig)
+                if (isinstance(self.dhcp_config, SegmentDhcpConfigV4) or
+                    isinstance(self.dhcp_config, SegmentDhcpConfigV6))
                 else self.dhcp_config)
+        elif (version.LooseVersion(nsx_version) >=
+              version.LooseVersion(nsx_constants.NSX_VERSION_3_0_0)):
+            body['dhcp_config'] = None
 
         return body
 
 
-class SegmentDhcpConfig(object):
+class SegmentDhcpConfigV4(object):
     def __init__(self, server_address=None, dns_servers=None,
-                 lease_time=None, options=None, is_ipv6=False):
-        if is_ipv6:
-            self.resource_type = 'SegmentDhcpV6Config'
-        else:
-            self.resource_type = 'SegmentDhcpV4Config'
-
+                 lease_time=None, options=None):
+        self.resource_type = 'SegmentDhcpV4Config'
         self.server_address = server_address
         self.dns_servers = dns_servers
         self.lease_time = lease_time
@@ -824,7 +832,37 @@ class SegmentDhcpConfig(object):
                 self.options.get_obj_dict()
                 if isinstance(self.options, DhcpOptions)
                 else self.options)
+        return body
 
+
+class SegmentDhcpConfig(SegmentDhcpConfigV4):
+    # alias to SegmentDhcpConfigV4, for backwards compatibility
+    def __init__(self, server_address=None, dns_servers=None,
+                 lease_time=None, options=None, is_ipv6=False):
+        super(SegmentDhcpConfig, self).__init__(
+            server_address=server_address, dns_servers=dns_servers,
+            lease_time=lease_time, options=options)
+
+
+class SegmentDhcpConfigV6(object):
+    def __init__(self, server_address=None, dns_servers=None,
+                 lease_time=None, domain_names=None):
+        self.resource_type = 'SegmentDhcpV6Config'
+        self.server_address = server_address
+        self.dns_servers = dns_servers
+        self.lease_time = lease_time
+        self.domain_names = domain_names
+
+    def get_obj_dict(self):
+        body = {'resource_type': self.resource_type}
+        if self.server_address:
+            body['server_address'] = self.server_address
+        if self.dns_servers:
+            body['dns_servers'] = self.dns_servers
+        if self.lease_time:
+            body['lease_time'] = self.lease_time
+        if self.domain_names:
+            body['domain_names'] = self.domain_names
         return body
 
 
@@ -866,7 +904,8 @@ class BaseSegmentDef(ResourceDef):
         if self.has_attr('subnets'):
             subnets = []
             if self.get_attr('subnets'):
-                subnets = [subnet.get_obj_dict()
+                subnets = [subnet.get_obj_dict(self.nsx_version)
+                           if isinstance(subnet, Subnet) else subnet
                            for subnet in self.get_attr('subnets')]
             self._set_attr_if_specified(body, 'subnets', value=subnets)
 
@@ -876,7 +915,8 @@ class BaseSegmentDef(ResourceDef):
             self._set_attr_if_specified(body, 'ip_pool_id',
                                         body_attr='advanced_config',
                                         value=adv_cfg)
-        self._set_attrs_if_specified(body, ['domain_name', 'vlan_ids'])
+        self._set_attrs_if_specified(
+            body, ['domain_name', 'vlan_ids', 'ls_id'])
         return body
 
     @staticmethod
@@ -884,6 +924,8 @@ class BaseSegmentDef(ResourceDef):
         return 'Segment'
 
     def _get_adv_config(self, ip_pool_id):
+        if ip_pool_id is None:
+            return {'address_pool_paths': []}
         ip_pool_def = IpPoolDef(ip_pool_id=ip_pool_id)
         ip_pool_path = ip_pool_def.get_resource_full_path()
         return {'address_pool_paths': [ip_pool_path]}
@@ -1019,7 +1061,6 @@ class DhcpV4StaticBindingConfig(ResourceDef):
 
     def get_obj_dict(self):
         body = super(DhcpV4StaticBindingConfig, self).get_obj_dict()
-        # TODO(asarfaty): add object or v4/6 options
         self._set_attrs_if_specified(body,
                                      ['gateway_address',
                                       'host_name',
@@ -1036,9 +1077,6 @@ class DhcpV6StaticBindingConfig(DhcpV4StaticBindingConfig):
     def resource_type():
         return 'DhcpV6StaticBindingConfig'
 
-    def path_defs(self):
-        return (TenantDef, SegmentDef)
-
     def get_obj_dict(self):
         body = super(DhcpV6StaticBindingConfig, self).get_obj_dict()
         self._set_attrs_if_specified(body,
@@ -1046,8 +1084,8 @@ class DhcpV6StaticBindingConfig(DhcpV4StaticBindingConfig):
                                       'dns_nameservers',
                                       'ip_addresses',
                                       'sntp_servers',
-                                      'preferred_time'])
-
+                                      'preferred_time',
+                                      'options'])
         return body
 
 
@@ -1122,12 +1160,30 @@ class SegmentPortDef(ResourceDef):
             self._set_attr_if_specified(body, 'admin_state',
                                         value=admin_state)
 
+        if (self.has_attr('init_state') and
+            self._version_dependant_attr_supported('init_state')):
+            valid_list = [nsx_constants.INIT_STATE_UNBLOCKED_VLAN,
+                          nsx_constants.INIT_STATE_RESTORE_VIF]
+            init_state = self.get_attr('init_state')
+            if init_state not in valid_list:
+                raise exceptions.InvalidInput(
+                    operation='create_segment_port',
+                    arg_val=init_state,
+                    arg_name='init_state')
+            self._set_attr_if_specified(body, 'init_state')
+
+        if (self.has_attr('extra_configs') and
+            self._version_dependant_attr_supported('extra_configs')):
+            self._set_attr_if_specified(body, 'extra_configs')
+
         return body
 
     @property
     def version_dependant_attr_map(self):
         return {'hyperbus_mode': nsx_constants.NSX_VERSION_3_0_0,
-                'admin_state': nsx_constants.NSX_VERSION_3_0_0}
+                'admin_state': nsx_constants.NSX_VERSION_3_0_0,
+                'init_state': nsx_constants.NSX_VERSION_3_1_0,
+                'extra_configs': nsx_constants.NSX_VERSION_3_0_2}
 
 
 class SegmentBindingMapDefBase(ResourceDef):
@@ -1180,6 +1236,75 @@ class SegmentSecProfilesBindingMapDef(SegmentBindingMapDefBase):
         return body
 
 
+class SegmentQosProfilesBindingMapDef(SegmentBindingMapDefBase):
+    @property
+    def path_pattern(self):
+        return (SEGMENTS_PATH_PATTERN +
+                "%s/segment-qos-profile-binding-maps/")
+
+    @staticmethod
+    def resource_type():
+        return 'SegmentQoSProfileBindingMap'
+
+    def get_obj_dict(self):
+        body = super(SegmentQosProfilesBindingMapDef, self).get_obj_dict()
+
+        if self.has_attr('segment_qos_profile_id'):
+            path = ""
+            if self.get_attr('segment_qos_profile_id'):
+                profile = QosProfileDef(
+                    profile_id=self.get_attr('segment_qos_profile_id'),
+                    tenant=self.get_tenant())
+                path = profile.get_resource_full_path()
+            self._set_attr_if_specified(
+                body, 'segment_qos_profile_id',
+                body_attr='segment_qos_profile_path',
+                value=path)
+
+        return body
+
+
+class SegmentDiscoveryProfilesBindingMapDef(SegmentBindingMapDefBase):
+    @property
+    def path_pattern(self):
+        return (SEGMENTS_PATH_PATTERN +
+                "%s/segment-discovery-profile-binding-maps/")
+
+    @staticmethod
+    def resource_type():
+        return 'SegmentDiscoveryProfileBindingMap'
+
+    def get_obj_dict(self):
+        body = super(SegmentDiscoveryProfilesBindingMapDef,
+                     self).get_obj_dict()
+
+        if self.has_attr('mac_discovery_profile_id'):
+            path = ""
+            if self.get_attr('mac_discovery_profile_id'):
+                profile = MacDiscoveryProfileDef(
+                    profile_id=self.get_attr('mac_discovery_profile_id'),
+                    tenant=self.get_tenant())
+                path = profile.get_resource_full_path()
+            self._set_attr_if_specified(
+                body, 'mac_discovery_profile_id',
+                body_attr='mac_discovery_profile_path',
+                value=path)
+
+        if self.has_attr('ip_discovery_profile_id'):
+            path = ""
+            if self.get_attr('ip_discovery_profile_id'):
+                profile = IpDiscoveryProfileDef(
+                    profile_id=self.get_attr('ip_discovery_profile_id'),
+                    tenant=self.get_tenant())
+                path = profile.get_resource_full_path()
+            self._set_attr_if_specified(
+                body, 'ip_discovery_profile_id',
+                body_attr='ip_discovery_profile_path',
+                value=path)
+
+        return body
+
+
 class SegmentPortBindingMapDefBase(ResourceDef):
 
     @property
@@ -1459,6 +1584,14 @@ class Condition(object):
                 'value': self.value,
                 'operator': self.operator}
 
+    def __eq__(self, other):
+        if isinstance(other, Condition):
+            return self.get_obj_dict() == other.get_obj_dict()
+        return False
+
+    def __hash__(self):
+        return hash(tuple(self.get_obj_dict().values()))
+
 
 class IPAddressExpression(object):
     def __init__(self, ip_addresses):
@@ -1495,6 +1628,46 @@ class NestedExpression(object):
         return {'resource_type': 'NestedExpression',
                 'expressions': [ex.get_obj_dict() for ex in self.expressions]}
 
+    def _get_unique_expressions(self):
+        """Only AND operator is supported in a nested expression.
+
+        When comparing two nested expressions, only checking the Conditions
+        in the expression list will suffice since all ConjunctionOperators
+        will be the same.
+        """
+        expr_list = self.get_obj_dict()['expressions']
+        unique_exprs = [expr for expr in expr_list
+                        if expr.get('resource_type') != 'ConjunctionOperator']
+        return unique_exprs
+
+    def __eq__(self, other):
+
+        def expr_identifier(expr):
+            return expr.get('member_type'), expr.get('value')
+
+        if not isinstance(other, NestedExpression):
+            return False
+        self_expr_list = self._get_unique_expressions()
+        other_expr_list = other._get_unique_expressions()
+        if len(self_expr_list) != len(other_expr_list):
+            return False
+        # Each expression dict in the list should be uniquely identified by
+        # (expr.member_type, expr.value pair). In case of segment expressions,
+        # the identifier will be of format ("Segment", "tag_scope|tag_value")
+        self_sorted = sorted(self_expr_list, key=expr_identifier)
+        other_sorted = sorted(other_expr_list, key=expr_identifier)
+        for i in range(len(self_sorted)):
+            if not all(item in other_sorted[i].items()
+                       for item in self_sorted[i].items()):
+                return False
+        return True
+
+    def __hash__(self):
+        # List is not hashable in python
+        value_list = [tuple(expr.values())
+                      for expr in self._get_unique_expressions()]
+        return hash(tuple(value_list))
+
 
 class GroupDef(ResourceDef):
 
@@ -1706,12 +1879,19 @@ class SecurityPolicyRuleBaseDef(ResourceDef):
     def get_obj_dict(self):
         body = super(SecurityPolicyRuleBaseDef, self).get_obj_dict()
         domain_id = self.get_attr('domain_id')
+
+        plain_groups = False
+        if self.has_attr('plain_groups'):
+            plain_groups = self.get_attr('plain_groups')
+
         if self.has_attr('source_groups'):
-            body['source_groups'] = self.get_groups_path(
-                domain_id, self.get_attr('source_groups'))
+            body['source_groups'] = (self.get_groups_path(
+                domain_id, self.get_attr('source_groups')) if not plain_groups
+                else (self.get_attr('source_groups') or [constants.ANY_GROUP]))
         if self.has_attr('dest_groups'):
-            body['destination_groups'] = self.get_groups_path(
-                domain_id, self.get_attr('dest_groups'))
+            body['destination_groups'] = (self.get_groups_path(
+                domain_id, self.get_attr('dest_groups')) if not plain_groups
+                else (self.get_attr('dest_groups') or [constants.ANY_GROUP]))
 
         self._set_attrs_if_specified(body, ['sequence_number', 'scope',
                                             'action', 'direction', 'logged',
@@ -2216,7 +2396,7 @@ class CertificateDef(ResourceDef):
 
     @staticmethod
     def resource_type():
-        return "TlsTrustData"
+        return "TlsCertificate"
 
     def get_obj_dict(self):
         body = super(CertificateDef, self).get_obj_dict()
@@ -2282,7 +2462,8 @@ class NsxPolicyApi(object):
     def partial_updates_supported(self):
         return self.partial_updates
 
-    def create_or_update(self, resource_def, partial_updates=False):
+    def create_or_update(self, resource_def, partial_updates=False,
+                         force=False):
         """Create or update a policy object.
 
         This api will update an existing object, or create a new one if it
@@ -2297,6 +2478,11 @@ class NsxPolicyApi(object):
         headers = None
         if partial_updates:
             headers = {'nsx-enable-partial-patch': 'true'}
+        if force:
+            if headers:
+                headers['X-Allow-Overwrite'] = 'true'
+            else:
+                headers = {'X-Allow-Overwrite': 'true'}
         self.client.patch(path, body, headers=headers)
 
     def create_with_parent(self, parent_def, resource_def):
@@ -2351,6 +2537,16 @@ class NsxPolicyApi(object):
         if entity:
             return entity['state']
 
+    def refresh_realized_state(self, path):
+        # Using POST, excepting 204
+        expected_results = [requests.codes.no_content]
+        return self.client.create(REALIZATION_REFRESH_PATH % path,
+                                  expected_results=expected_results)
+
+    def get_intent_consolidated_status(self, path, silent=False):
+        return self.client.get(REALIZATION_STATUS_PATH % path,
+                               silent=silent)
+
 
 class RouteMapEntry(object):
     def __init__(self, action, community_list_matches=None,
