@@ -15,6 +15,7 @@
 #
 
 import abc
+import copy
 
 from oslo_log import log as logging
 import six
@@ -30,11 +31,6 @@ from vmware_nsxlib.v3 import utils
 
 LOG = logging.getLogger(__name__)
 
-# Sentitel object to indicate unspecified attribute value
-# None value in attribute would indicate "unset" functionality,
-# while "ignore" means that the value not be present in request
-# body
-
 
 class NsxPolicyLBAppProfileBase(NsxPolicyResourceBase):
     """NSX Policy LB app profile"""
@@ -77,14 +73,14 @@ class NsxPolicyLBAppProfileBase(NsxPolicyResourceBase):
         lb_app_profile_def = self.entry_def(
             lb_app_profile_id=lb_app_profile_id,
             tenant=tenant)
-        self.policy_api.delete(lb_app_profile_def)
+        self._delete_with_retry(lb_app_profile_def)
 
     def get(self, lb_app_profile_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lb_app_profile_def = self.entry_def(
             lb_app_profile_id=lb_app_profile_id,
             tenant=tenant)
-        return self.policy_api.get(lb_app_profile_def)
+        return self.policy_api.get(lb_app_profile_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lb_app_profile_def = self.entry_def(tenant=tenant)
@@ -182,14 +178,14 @@ class NsxPolicyLoadBalancerServerSSLProfileApi(NsxPolicyResourceBase):
         lb_server_ssl_profile_def = self.entry_def(
             server_ssl_profile_id=server_ssl_profile_id,
             tenant=tenant)
-        self.policy_api.delete(lb_server_ssl_profile_def)
+        self._delete_with_retry(lb_server_ssl_profile_def)
 
     def get(self, server_ssl_profile_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lb_server_ssl_profile_def = self.entry_def(
             server_ssl_profile_id=server_ssl_profile_id,
             tenant=tenant)
-        return self.policy_api.get(lb_server_ssl_profile_def)
+        return self.policy_api.get(lb_server_ssl_profile_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lb_server_ssl_profile_def = self.entry_def(tenant=tenant)
@@ -239,14 +235,14 @@ class NsxPolicyLoadBalancerClientSSLProfileApi(NsxPolicyResourceBase):
         lb_client_ssl_profile_def = self.entry_def(
             client_ssl_profile_id=client_ssl_profile_id,
             tenant=tenant)
-        self.policy_api.delete(lb_client_ssl_profile_def)
+        self._delete_with_retry(lb_client_ssl_profile_def)
 
     def get(self, client_ssl_profile_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lb_client_ssl_profile_def = self.entry_def(
             client_ssl_profile_id=client_ssl_profile_id,
             tenant=tenant)
-        return self.policy_api.get(lb_client_ssl_profile_def)
+        return self.policy_api.get(lb_client_ssl_profile_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lb_client_ssl_profile_def = self.entry_def(tenant=tenant)
@@ -299,14 +295,14 @@ class NsxPolicyLoadBalancerPersistenceProfileApi(
         persistence_profile_def = self.entry_def(
             persistence_profile_id=persistence_profile_id,
             tenant=tenant)
-        self.policy_api.delete(persistence_profile_def)
+        self._delete_with_retry(persistence_profile_def)
 
     def get(self, persistence_profile_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         persistence_profile_def = self.entry_def(
             persistence_profile_id=persistence_profile_id,
             tenant=tenant)
-        return self.policy_api.get(persistence_profile_def)
+        return self.policy_api.get(persistence_profile_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         persistence_profile_def = self.entry_def(tenant=tenant)
@@ -501,13 +497,13 @@ class NsxPolicyLoadBalancerPoolApi(NsxPolicyResourceBase):
                tenant=constants.POLICY_INFRA_TENANT):
         lb_pool_def = self.entry_def(
             lb_pool_id=lb_pool_id, tenant=tenant)
-        self.policy_api.delete(lb_pool_def)
+        self._delete_with_retry(lb_pool_def)
 
     def get(self, lb_pool_id, tenant=constants.POLICY_INFRA_TENANT,
             silent=False):
         lb_pool_def = self.entry_def(
             lb_pool_id=lb_pool_id, tenant=tenant)
-        return self.policy_api.get(lb_pool_def)
+        return self.policy_api.get(lb_pool_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lb_pool_def = self.entry_def(tenant=tenant)
@@ -517,7 +513,8 @@ class NsxPolicyLoadBalancerPoolApi(NsxPolicyResourceBase):
                tags=IGNORE, members=IGNORE, algorithm=IGNORE,
                active_monitor_paths=IGNORE, member_group=IGNORE,
                snat_translation=IGNORE,
-               tenant=constants.POLICY_INFRA_TENANT):
+               tenant=constants.POLICY_INFRA_TENANT,
+               allow_partial_updates=True):
         self._update(
             lb_pool_id=lb_pool_id,
             name=name,
@@ -528,7 +525,7 @@ class NsxPolicyLoadBalancerPoolApi(NsxPolicyResourceBase):
             algorithm=algorithm,
             member_group=member_group,
             snat_translation=snat_translation,
-            tenant=tenant)
+            tenant=tenant, allow_partial_updates=allow_partial_updates)
 
     def add_monitor_to_pool(self, lb_pool_id, active_monitor_paths,
                             tenant=constants.POLICY_INFRA_TENANT):
@@ -662,16 +659,26 @@ class NsxPolicyLoadBalancerServiceApi(NsxPolicyResourceBase):
                tenant=constants.POLICY_INFRA_TENANT):
         lb_service_def = self.entry_def(
             lb_service_id=lb_service_id, tenant=tenant)
-        self.policy_api.delete(lb_service_def)
+        self._delete_with_retry(lb_service_def)
 
-    def get(self, lb_service_id, tenant=constants.POLICY_INFRA_TENANT):
+    def get(self, lb_service_id,
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lb_service_def = self.entry_def(
             lb_service_id=lb_service_id, tenant=tenant)
-        return self.policy_api.get(lb_service_def)
+        return self.policy_api.get(lb_service_def, silent=silent)
 
-    def list(self, tenant=constants.POLICY_INFRA_TENANT):
+    def list(self, tenant=constants.POLICY_INFRA_TENANT, silent=False,
+             silent_if_empty=False):
+        if not silent and silent_if_empty:
+            # Log only non-empty results
+            silent = True
         lb_service_def = lb_defs.LBServiceDef(tenant=tenant)
-        return self.policy_api.list(lb_service_def)['results']
+        list_results = self.policy_api.list(lb_service_def, silent=silent)
+        if silent_if_empty and list_results.get('results'):
+            LOG.debug("REST call: GET %s. Response: %s",
+                      lb_service_def.get_section_path(), list_results)
+
+        return list_results['results']
 
     def update(self, lb_service_id, name=IGNORE,
                description=IGNORE, tags=IGNORE,
@@ -689,21 +696,47 @@ class NsxPolicyLoadBalancerServiceApi(NsxPolicyResourceBase):
             relax_scale_validation=relax_scale_validation,
             tenant=tenant)
 
+    def update_customized(self, lb_service_id, update_payload_cbk,
+                          tenant=constants.POLICY_INFRA_TENANT):
+        """Update the LB service using GET & PUT
+
+        Changing the body with a customized callback
+        """
+
+        lb_service_def = self.entry_def(
+            lb_service_id=lb_service_id, tenant=tenant)
+        lb_service_path = lb_service_def.get_resource_path()
+
+        @utils.retry_upon_exception(
+            nsxlib_exc.StaleRevision,
+            max_attempts=self.policy_api.client.max_attempts)
+        def _update():
+            # Get the current data of service
+            lb_service_body = self.policy_api.get(lb_service_def)
+            # Update the body with the supplied callback
+            update_payload_cbk(lb_service_body)
+            # Update the backend using PUT
+            self.policy_api.client.update(lb_service_path, lb_service_body)
+
+        _update()
+
     def get_statistics(self, lb_service_id,
-                       tenant=constants.POLICY_INFRA_TENANT):
+                       tenant=constants.POLICY_INFRA_TENANT,
+                       silent=False):
         lb_service_stats_def = (
             lb_defs.LBServiceStatisticsDef(
                 lb_service_id=lb_service_id,
                 tenant=tenant))
-        return self.policy_api.get(lb_service_stats_def)
+        return self.policy_api.get(lb_service_stats_def, silent=silent)
 
     def get_status(self, lb_service_id,
-                   tenant=constants.POLICY_INFRA_TENANT):
+                   tenant=constants.POLICY_INFRA_TENANT,
+                   silent=False):
         lb_service_status_def = (
             lb_defs.LBServiceStatusDef(
                 lb_service_id=lb_service_id,
                 tenant=tenant))
-        return self.policy_api.get(lb_service_status_def)
+        return self.policy_api.get(lb_service_status_def, silent=silent)
 
     def get_virtual_server_status(self, lb_service_id, lb_virtual_server_id,
                                   tenant=constants.POLICY_INFRA_TENANT):
@@ -785,13 +818,13 @@ class NsxPolicyLoadBalancerVirtualServerAPI(NsxPolicyResourceBase):
                tenant=constants.POLICY_INFRA_TENANT):
         lbvs_def = self.entry_def(
             virtual_server_id=virtual_server_id, tenant=tenant)
-        self.policy_api.delete(lbvs_def)
+        self._delete_with_retry(lbvs_def)
 
     def get(self, virtual_server_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lbvs_def = self.entry_def(
             virtual_server_id=virtual_server_id, tenant=tenant)
-        return self.policy_api.get(lbvs_def)
+        return self.policy_api.get(lbvs_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lbvs_def = self.entry_def(tenant=tenant)
@@ -881,6 +914,17 @@ class NsxPolicyLoadBalancerVirtualServerAPI(NsxPolicyResourceBase):
         self.policy_api.create_or_update(
             lbvs_def, partial_updates=False)
 
+    def remove_dlb_virtual_server_persistence_profile(
+            self, virtual_server_id, tenant=constants.POLICY_INFRA_TENANT):
+        dlb_vs_def = self._get_and_update_def(
+            virtual_server_id=virtual_server_id, tenant=tenant)
+        body = dlb_vs_def.body if dlb_vs_def.body else {}
+        body.pop('lb_persistence_profile_path', None)
+        if body:
+            dlb_vs_def.set_obj_dict(body)
+        self.policy_api.create_or_update(
+            dlb_vs_def, partial_updates=False)
+
     def update_virtual_server_with_vip(self, virtual_server_id, vip,
                                        tenant=constants.POLICY_INFRA_TENANT):
         return self.update(
@@ -942,9 +986,29 @@ class NsxPolicyLoadBalancerVirtualServerAPI(NsxPolicyResourceBase):
                             vs_data=body,
                             rules=lb_rules, tenant=tenant)
 
+    def update_lb_rules(self, virtual_server_id, rules,
+                        tenant=constants.POLICY_INFRA_TENANT):
+        lbvs_def = self.entry_def(
+            virtual_server_id=virtual_server_id,
+            tenant=tenant)
+        lbvs_path = lbvs_def.get_resource_path()
+
+        @utils.retry_upon_exception(
+            nsxlib_exc.StaleRevision,
+            max_attempts=self.policy_api.client.max_attempts)
+        def _update():
+            # Get the current data of vs
+            lbvs_body = self.policy_api.get(lbvs_def)
+            lbvs_body['rules'] = copy.deepcopy(rules)
+            # Update the backend using PUT
+            self.policy_api.client.update(lbvs_path, lbvs_body)
+
+        _update()
+
     def update_lb_rule(self, virtual_server_id, lb_rule_name,
                        actions=None, match_conditions=None,
                        match_strategy=None, phase=None, position=-1,
+                       compare_name_suffix=None,
                        tenant=constants.POLICY_INFRA_TENANT):
         lb_rule = lb_defs.LBRuleDef(
             actions, match_conditions, lb_rule_name, match_strategy, phase)
@@ -956,8 +1020,13 @@ class NsxPolicyLoadBalancerVirtualServerAPI(NsxPolicyResourceBase):
 
         # Remove existing rule
         try:
-            rule_index = next(lb_rules.index(r) for r in lb_rules
-                              if r.get('display_name') == lb_rule_name)
+            if compare_name_suffix:
+                rule_index = next(lb_rules.index(r) for r in lb_rules
+                                  if r.get('display_name',
+                                           '').endswith(compare_name_suffix))
+            else:
+                rule_index = next(lb_rules.index(r) for r in lb_rules
+                                  if r.get('display_name') == lb_rule_name)
         except Exception:
             err_msg = (_("No resource in rules matched for values: "
                          "%(values)s") % {'values': lb_rule_name})
@@ -976,13 +1045,19 @@ class NsxPolicyLoadBalancerVirtualServerAPI(NsxPolicyResourceBase):
             rules=lb_rules, vs_data=body, tenant=tenant)
 
     def remove_lb_rule(self, virtual_server_id, lb_rule_name,
+                       check_name_suffix=False,
                        tenant=constants.POLICY_INFRA_TENANT):
         lbvs_def = self.entry_def(virtual_server_id=virtual_server_id,
                                   tenant=tenant)
         body = self.policy_api.get(lbvs_def)
         lb_rules = body.get('rules', [])
-        lb_rules = [r for r in lb_rules if (r.get('display_name') !=
-                                            lb_rule_name)]
+        if check_name_suffix:
+            lb_rules = [r for r in lb_rules
+                        if not r.get('display_name', '').endswith(
+                            lb_rule_name)]
+        else:
+            lb_rules = [r for r in lb_rules
+                        if r.get('display_name') != lb_rule_name]
         return self._update(
             virtual_server_id=virtual_server_id, vs_data=body,
             rules=lb_rules, tenant=tenant)
@@ -1030,14 +1105,14 @@ class NsxPolicyLBMonitorProfileBase(NsxPolicyResourceBase):
         lb_monitor_profile_def = self.entry_def(
             lb_monitor_profile_id=lb_monitor_profile_id,
             tenant=tenant)
-        self.policy_api.delete(lb_monitor_profile_def)
+        self._delete_with_retry(lb_monitor_profile_def)
 
     def get(self, lb_monitor_profile_id,
-            tenant=constants.POLICY_INFRA_TENANT):
+            tenant=constants.POLICY_INFRA_TENANT, silent=False):
         lb_monitor_profile_def = self.entry_def(
             lb_monitor_profile_id=lb_monitor_profile_id,
             tenant=tenant)
-        return self.policy_api.get(lb_monitor_profile_def)
+        return self.policy_api.get(lb_monitor_profile_def, silent=silent)
 
     def list(self, tenant=constants.POLICY_INFRA_TENANT):
         lb_monitor_profile_def = self.entry_def(tenant=tenant)
