@@ -17,6 +17,7 @@ import abc
 import collections
 import inspect
 import re
+from threading import Lock
 import time
 
 from oslo_log import log
@@ -25,7 +26,7 @@ from tenacity import _utils as tenacity_utils
 
 from vmware_nsxlib._i18n import _
 from vmware_nsxlib.v3 import constants
-from vmware_nsxlib.v3 import exceptions as nsxlib_exceptions
+from vmware_nsxlib.v3 import exceptions as nsxlib_exc
 from vmware_nsxlib.v3 import nsx_constants
 
 LOG = log.getLogger(__name__)
@@ -63,7 +64,7 @@ def set_inject_headers_callback(callback):
 
 
 def censor_headers(headers):
-    censored_headers = ['authorization']
+    censored_headers = ['authorization', 'x-xsrf-token', 'cookie']
     result = {}
     for name, value in headers.items():
         if name.lower() in censored_headers:
@@ -104,7 +105,7 @@ def update_tag_limits(limits):
 def _validate_resource_type_length(resource_type):
     # Add in a validation to ensure that we catch this at build time
     if len(resource_type) > MAX_RESOURCE_TYPE_LEN:
-        raise nsxlib_exceptions.NsxLibInvalidInput(
+        raise nsxlib_exc.NsxLibInvalidInput(
             error_message=(_('Resource type cannot exceed %(max_len)s '
                              'characters: %(resource_type)s') %
                            {'max_len': MAX_RESOURCE_TYPE_LEN,
@@ -180,6 +181,7 @@ def _log_after_retry(retry_state):
 
 def retry_upon_exception(exc, delay=0.5, max_delay=2,
                          max_attempts=DEFAULT_MAX_ATTEMPTS):
+    # exc can be a single exception or a tuple of exceptions
     return tenacity.retry(reraise=True,
                           retry=tenacity.retry_if_exception_type(exc),
                           wait=tenacity.wait_exponential(
@@ -213,6 +215,26 @@ def retry_upon_none_result(max_attempts, delay=0.5, max_delay=10,
                           before=_log_before_retry, after=_log_after_retry)
 
 
+class RetryAttemptsExceeded(tenacity.RetryError):
+    def reraise(self):
+        raise self.last_attempt.result()
+
+
+# Retry when exception is returned by decorated function.
+# If retry attempts are exceeded, reraise the last exception.
+# This is achieved by overriding reraise method of RetryAttemptsExceeded
+def retry_random_upon_exception_result(max_attempts, delay=0.5, max_delay=10):
+    wait_func = tenacity.wait_random_exponential(
+        multiplier=delay, max=max_delay)
+    return tenacity.retry(reraise=True,
+                          retry_error_cls=RetryAttemptsExceeded,
+                          retry=tenacity.retry_if_result(
+                              lambda x: isinstance(x, Exception)),
+                          wait=wait_func,
+                          stop=tenacity.stop_after_attempt(max_attempts),
+                          before=_log_before_retry, after=_log_after_retry)
+
+
 def list_match(list1, list2):
     # Check if list1 and list2 have identical elements, but relaxed on
     # dict elements where list1's dict element can be a subset of list2's
@@ -253,8 +275,12 @@ def dict_match(dict1, dict2):
     return True
 
 
+def get_name_short_uuid(uuid):
+    return '_' + uuid[:5] + '...' + uuid[-5:]
+
+
 def get_name_and_uuid(name, uuid, tag=None, maxlen=80):
-    short_uuid = '_' + uuid[:5] + '...' + uuid[-5:]
+    short_uuid = get_name_short_uuid(uuid)
     maxlen = maxlen - len(short_uuid)
     if tag:
         maxlen = maxlen - len(tag) - 1
@@ -318,6 +344,8 @@ class NsxLibApiBase(object):
         self.nsxlib = nsxlib
         super(NsxLibApiBase, self).__init__()
         self.cache = NsxLibCache(self.cache_timeout)
+        self.max_attempts = (self.client.max_attempts
+                             if hasattr(self.client, 'max_attempts') else 1)
 
     @abc.abstractproperty
     def uri_segment(self):
@@ -367,7 +395,7 @@ class NsxLibApiBase(object):
     def delete(self, uuid):
         if self.use_cache_for_get:
             self.cache.remove(uuid)
-        return self.client.delete(self.get_path(uuid))
+        return self._delete_with_retry(uuid)
 
     def find_by_display_name(self, display_name):
         found = []
@@ -428,9 +456,8 @@ class NsxLibApiBase(object):
             # NSX has, we will get a 412: Precondition Failed.
             # In that case we need to re-fetch, patch the response and send
             # it again with the new revision_id
-            @retry_upon_exception(
-                nsxlib_exceptions.StaleRevision,
-                max_attempts=self.client.max_attempts)
+            @retry_upon_exception(nsxlib_exc.StaleRevision,
+                                  max_attempts=self.max_attempts)
             def do_update():
                 return self._internal_update_resource(
                     resource, payload,
@@ -450,21 +477,22 @@ class NsxLibApiBase(object):
                 action_params=action_params,
                 update_payload_cbk=update_payload_cbk)
 
-    def _delete_with_retry(self, resource):
+    def _delete_with_retry(self, resource, headers=None):
+        self._delete_by_path_with_retry(self.get_path(resource), headers)
+
+    def _delete_by_path_with_retry(self, path, headers=None):
         # Using internal method so we can access max_attempts in the decorator
-        @retry_upon_exception(
-            nsxlib_exceptions.StaleRevision,
-            max_attempts=self.client.max_attempts)
+        @retry_upon_exception(nsxlib_exc.StaleRevision,
+                              max_attempts=self.max_attempts)
         def _do_delete():
-            self.client.delete(self.get_path(resource))
+            self.client.delete(path, headers=headers)
 
         _do_delete()
 
     def _create_with_retry(self, resource, body=None, headers=None):
         # Using internal method so we can access max_attempts in the decorator
-        @retry_upon_exception(
-            nsxlib_exceptions.StaleRevision,
-            max_attempts=self.client.max_attempts)
+        @retry_upon_exception(nsxlib_exc.StaleRevision,
+                              max_attempts=self.max_attempts)
         def _do_create():
             return self.client.create(resource, body, headers=headers)
 
@@ -485,11 +513,11 @@ class NsxLibApiBase(object):
         if len(matched_results) == 0:
             err_msg = (_("Could not find %(resource)s %(name)s") %
                        {'name': name_or_id, 'resource': resource})
-            raise nsxlib_exceptions.ManagerError(details=err_msg)
+            raise nsxlib_exc.ManagerError(details=err_msg)
         elif len(matched_results) > 1:
             err_msg = (_("Found multiple %(resource)s named %(name)s") %
                        {'name': name_or_id, 'resource': resource})
-            raise nsxlib_exceptions.ManagerError(details=err_msg)
+            raise nsxlib_exc.ManagerError(details=err_msg)
 
         return matched_results[0].get('id')
 
@@ -572,27 +600,27 @@ def validate_icmp_params(icmp_type, icmp_code, icmp_version=4, strict=False):
     if icmp_version != 4:
         # ICMPv6 is currently not supported
         return
-    if icmp_type:
+    if icmp_type is not None:
         if (strict and icmp_type not in
                 constants.IPV4_ICMP_STRICT_TYPES):
-            raise nsxlib_exceptions.InvalidInput(
+            raise nsxlib_exc.InvalidInput(
                 operation='create_rule',
                 arg_val=icmp_type,
                 arg_name='icmp_type')
         if icmp_type not in constants.IPV4_ICMP_TYPES:
-            raise nsxlib_exceptions.InvalidInput(
+            raise nsxlib_exc.InvalidInput(
                 operation='create_rule',
                 arg_val=icmp_type,
                 arg_name='icmp_type')
         if (icmp_code and strict and icmp_code not in
                 constants.IPV4_ICMP_STRICT_TYPES[icmp_type]):
-            raise nsxlib_exceptions.InvalidInput(
+            raise nsxlib_exc.InvalidInput(
                 operation='create_rule',
                 arg_val=icmp_code,
                 arg_name='icmp_code for this icmp_type')
         if (icmp_code and icmp_code not in
                 constants.IPV4_ICMP_TYPES[icmp_type]):
-            raise nsxlib_exceptions.InvalidInput(
+            raise nsxlib_exc.InvalidInput(
                 operation='create_rule',
                 arg_val=icmp_code,
                 arg_name='icmp_code for this icmp_type')
@@ -606,7 +634,7 @@ def get_l4_protocol_name(protocol_number):
     try:
         protocol_number = int(protocol_number)
     except ValueError:
-        raise nsxlib_exceptions.InvalidInput(
+        raise nsxlib_exc.InvalidInput(
             operation='create_rule',
             arg_val=protocol_number,
             arg_name='protocol')
@@ -664,3 +692,111 @@ def get_dhcp_opt_code(name):
         'reboot-time': 211,
     }
     return _supported_options.get(name)
+
+
+class APIRateLimiter(object):
+    def __init__(self, max_calls, period=1.0):
+        self._enabled = max_calls is not None
+        if not self._enabled:
+            return
+        if period <= 0 or int(max_calls) <= 0:
+            raise ValueError('period and max_calls should be positive')
+        self._period = period
+        self._max_calls = min(int(max_calls),
+                              constants.API_DEFAULT_MAX_RATE)
+        self._call_time = collections.deque()
+        self._lock = Lock()
+
+    def __enter__(self):
+        if not self._enabled:
+            return 0
+        pre_wait_ts = time.time()
+        with self._lock:
+            wait_time = self._calc_wait_time()
+            if wait_time:
+                time.sleep(wait_time)
+            # assume api call happens immediately after entering context
+            post_wait_ts = time.time()
+            self._call_time.append(post_wait_ts)
+            return post_wait_ts - pre_wait_ts
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        pass
+
+    def adjust_rate(self, **kwargs):
+        pass
+
+    def _calc_wait_time(self):
+        now = time.time()
+        # remove timestamps out of current window
+        while self._call_time and now - self._period > self._call_time[0]:
+            self._call_time.popleft()
+        current_rate = len(self._call_time)
+        if current_rate < self._max_calls:
+            return 0
+        # call_time contains at least #max_rate timestamps.
+        # earliest possible time to get below rate limit is at
+        # T = self.call_time[-self.max_calls] + self.period
+        # Thus need to wait T - now
+        return self._call_time[-self._max_calls] + self._period - now
+
+
+class APIRateLimiterAIMD(APIRateLimiter):
+    def __init__(self, max_calls, period=1.0):
+        super(APIRateLimiterAIMD, self).__init__(max_calls, period=period)
+        if self._enabled:
+            self._top_rate = self._max_calls
+            self._max_calls = 1
+            self._pos_sig = 0
+            self._neg_sig = 0
+            self._last_adjust_rate = time.time()
+
+    def adjust_rate(self, wait_time=0.0, status_code=200, **kwargs):
+        if not self._enabled:
+            return
+        with self._lock:
+            if status_code in constants.API_REDUCE_RATE_CODES:
+                self._neg_sig += 1
+            elif wait_time >= constants.API_WAIT_MIN_THRESHOLD:
+                self._pos_sig += 1
+
+            now = time.time()
+            if now - self._last_adjust_rate >= self._period:
+                if self._neg_sig > 0:
+                    self._max_calls = max(self._max_calls // 2, 1)
+                    LOG.debug("Decreasing API rate limit to %d due to HTTP "
+                              "status code %d", self._max_calls, status_code)
+                elif self._pos_sig > 0:
+                    self._max_calls = min(self._max_calls + 1, self._top_rate)
+                    LOG.debug("Increasing API rate limit to %d with HTTP "
+                              "status code %d", self._max_calls, status_code)
+                self._pos_sig = 0
+                self._neg_sig = 0
+                self._last_adjust_rate = now
+
+
+class APICallRecord(object):
+    def __init__(self, verb, uri, status, provider=None, timestamp=None):
+        self.timestamp = timestamp or time.time()
+        self.verb = verb
+        self.uri = uri
+        self.status = status
+        self.provider = provider
+
+
+class APICallCollector(object):
+    def __init__(self, provider, max_entry=50000):
+        self._api_log_store = collections.deque(maxlen=max_entry)
+        self.provider = provider
+
+    def add_record(self, record):
+        self._api_log_store.append(record)
+
+    def pop_record(self):
+        return self._api_log_store.popleft()
+
+    def pop_all_records(self):
+        records = []
+        while len(self._api_log_store) > 0:
+            records.append(self.pop_record())
+        return records
